import logging

from mayan.apps.common.utils import return_attrib

from .literals import EVENT_MANAGER_ORDER_AFTER

logger = logging.getLogger(name=__name__)


class EventManager:
    """
    keep_attributes - List of event related object attributes that should
    not be removed after the event is committed.
    """
    EVENT_ATTRIBUTES = ('ignore', 'keep_attributes', 'type')
    EVENT_ARGUMENTS = ('actor', 'action_object', 'target')

    def __init__(self, instance, **kwargs):
        self.instance = instance
        self.instance_event_attributes = {}
        self.kwargs = kwargs

    def commit(self):
        if not self.instance_event_attributes['ignore']:
            self._commit()
        else:
            # If the event is ignored, restore the event related attributes
            # that were removed via .pop().
            for key, value in self.instance_event_attributes.items():
                if key not in ('ignore', 'type'):
                    setattr(
                        self.instance, '_event_{}'.format(key), value
                    )

    def get_event_arguments(self, argument_map):
        result = {}

        for argument in self.EVENT_ARGUMENTS:
            # Grab the static argument value from the argument map.
            # If the argument is not in the map, it is dynamic and must be
            # obtained from the instance attributes.
            value = argument_map.get(
                argument, self.instance_event_attributes[argument]
            )

            if value == 'self':
                result[argument] = self.instance
            elif isinstance(value, str):
                result[argument] = return_attrib(
                    attrib=value, obj=self.instance
                )
            else:
                result[argument] = value

        return result

    def pop_event_attributes(self):
        for attribute in self.EVENT_ATTRIBUTES:
            # If the attribute is not set or is set but is None.
            if not self.instance_event_attributes.get(attribute, None):
                full_name = '_event_{}'.format(attribute)
                value = self.instance.__dict__.pop(full_name, None)
                self.instance_event_attributes[attribute] = value

        keep_attributes = self.instance_event_attributes['keep_attributes'] or ()

        # Allow passing a runtime defined event.
        if self.instance_event_attributes['type']:
            self.kwargs['event'] = self.instance_event_attributes['type']

        for attribute in self.EVENT_ARGUMENTS:
            # If the attribute is not set or is set but is None.
            if not self.instance_event_attributes.get(attribute, None):
                full_name = '_event_{}'.format(attribute)
                if full_name in keep_attributes:
                    value = self.instance.__dict__.get(full_name, None)
                else:
                    value = self.instance.__dict__.pop(full_name, None)

                self.instance_event_attributes[attribute] = value

    def prepare(self):
        """Optional method to gather information before the actual commit."""


class EventManagerMethodAfter(EventManager):
    order = EVENT_MANAGER_ORDER_AFTER

    def _commit(self):
        kwargs = self.get_event_arguments(argument_map=self.kwargs)
        self.kwargs['event'].commit(**kwargs)


class EventManagerSave(EventManager):
    order = EVENT_MANAGER_ORDER_AFTER

    def _commit(self):
        if self.created:
            if 'created' in self.kwargs:
                kwargs = self.get_event_arguments(
                    argument_map=self.kwargs['created']
                )
                self.kwargs['created']['event'].commit(**kwargs)
        else:
            if 'edited' in self.kwargs:
                kwargs = self.get_event_arguments(
                    argument_map=self.kwargs['edited']
                )
                self.kwargs['edited']['event'].commit(**kwargs)

    def prepare(self):
        self.created = not self.instance.pk



from django.db import connection, migrations
from django.db.backends.ddl_references import IndexName, Statement, Table
from django.db.migrations.loader import MigrationLoader


def code_drop_btree_index(apps, schema_editor):
    """
    Process BTREE indexes that were not renamed along with their respective
    models.
    """
    index_hashes = ('30bada95', '42757b7a')

    # Access the model from the previous migration to allow deleting the
    # indexes.
    loader = MigrationLoader(connection=connection)
    state = loader.project_state(
        nodes=('documents', '0056_auto_20200916_0959')
    )

    DocumentVersion = state.apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    DocumentVersion._meta.db_table = 'documents_documentfile'

    for index_name in schema_editor._constraint_names(model=DocumentVersion):
        for index_hash in index_hashes:
            if index_hash in index_name:
                if '_fk_' not in index_name:
                    # Index is needed in a foreign key constraint as it
                    # happens with MySQL. Ignore dropping the index.
                    schema_editor.execute(
                        sql=schema_editor._delete_index_sql(
                            model=DocumentVersion, name=index_name
                        )
                    )


def code_drop_btree_index_reverse(apps, schema_editor):
    """
    Create the BTREE indexes that were not renamed along with their
    respective models.
    """
    loader = MigrationLoader(connection=connection)
    state = loader.project_state(
        nodes=('documents', '0056_auto_20200916_0959')
    )

    DocumentVersion = state.apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )

    field_names = ('document_id', 'timestamp')

    for field_name in field_names:
        fields = (
            DocumentVersion._meta.get_field(field_name=field_name),  # Comma as `fields` is expecting an iterable.
        )

        tablespace_sql = schema_editor._get_index_tablespace_sql(
            model=DocumentVersion, fields=fields, db_tablespace=None
        )

        columns = [field.column for field in fields]
        table = DocumentVersion._meta.db_table

        def create_index_name(*args, **kwargs):
            name = schema_editor._create_index_name(*args, **kwargs)
            return schema_editor.quote_name(name)

        condition = None

        sql = Statement(
            template=schema_editor.sql_create_index,
            table=Table(
                table='documents_documentfile',
                quote_name=schema_editor.quote_name
            ),
            name=IndexName(
                table=table, columns=columns, suffix='',
                create_index_name=create_index_name
            ),
            using='',
            columns=schema_editor._index_columns(
                table=table, columns=columns, col_suffixes=(), opclasses=()
            ),
            extra=tablespace_sql,
            condition=(' WHERE ' + condition) if condition else '',
        )
        schema_editor.execute(sql=sql)


class RunPythonAtomicDynamic(migrations.RunPython):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if connection.features.can_rollback_ddl:
            self.atomic = True
        else:
            self.atomic = False


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0056_auto_20200916_0959')
    ]

    operations = [
        migrations.RenameModel(
            old_name='DocumentVersion',
            new_name='DocumentFile'
        ),
        RunPythonAtomicDynamic(
            code=code_drop_btree_index,
            reverse_code=code_drop_btree_index_reverse
        ),
        migrations.AlterModelOptions(
            name='documentpageresult',
            options={
                'ordering': (
                    'document_file__document', 'page_number'
                ),
                'verbose_name': 'Document page',
                'verbose_name_plural': 'Document pages'
            }
        ),
        migrations.RenameField(
            model_name='documentpage',
            old_name='document_version',
            new_name='document_file'
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Document exports'), name='document_exports'
)

event_document_version_exported = namespace.add_event_type(
    label=_(message='Document version exported'), name='document_version_exported'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.converter.permissions import (
    permission_transformation_delete, permission_transformation_edit
)
from mayan.apps.navigation.classes import Link

from ..icons import (
    icon_document_file_delete, icon_document_file_edit,
    icon_document_file_introspect, icon_document_file_list,
    icon_document_file_preview, icon_document_file_print,
    icon_document_file_properties_detail, icon_document_file_return_list,
    icon_document_file_return_to_document,
    icon_document_file_transformation_list_clear,
    icon_document_file_transformation_list_clone
)
from ..permissions import (
    permission_document_file_delete, permission_document_file_edit,
    permission_document_file_print, permission_document_file_tools,
    permission_document_file_view, permission_document_view
)

link_document_file_delete = Link(
    args='object.pk',
    icon=icon_document_file_delete,
    permission=permission_document_file_delete, tags='dangerous',
    text=_(message='Delete'), view='documents:document_file_delete',
)
link_document_file_multiple_delete = Link(
    icon=icon_document_file_delete,
    permission=permission_document_file_delete, tags='dangerous',
    text=_(message='Delete'), view='documents:document_file_multiple_delete',
)
link_document_file_edit = Link(
    args='object.pk', icon=icon_document_file_edit,
    permission=permission_document_file_edit,
    text=_(message='Edit'), view='documents:document_file_edit',
)
link_document_file_introspect_multiple = Link(
    icon=icon_document_file_introspect,
    text=_(message='Introspect'),
    view='documents:document_file_introspect_multiple'
)
link_document_file_introspect_single = Link(
    args='resolved_object.pk',
    icon=icon_document_file_introspect,
    permission=permission_document_file_tools,
    text=_(message='Introspect'),
    view='documents:document_file_introspect_single'
)
link_document_file_list = Link(
    args='resolved_object.pk',
    icon=icon_document_file_list,
    permission=permission_document_file_view,
    text=_(message='Files'), view='documents:document_file_list',
)
link_document_file_print_form = Link(
    args='resolved_object.id', icon=icon_document_file_print,
    permission=permission_document_file_print, text=_(message='Print'),
    view='documents:document_file_print_form'
)
link_document_file_properties = Link(
    args='resolved_object.id',
    icon=icon_document_file_properties_detail,
    permission=permission_document_file_view,
    text=_(message='Properties'), view='documents:document_file_properties',
)
link_document_file_return_to_document = Link(
    args='resolved_object.document.pk',
    icon=icon_document_file_return_to_document,
    permission=permission_document_view, text=_(message='Document'),
    view='documents:document_preview',
)
link_document_file_return_list = Link(
    args='resolved_object.document.pk',
    icon=icon_document_file_return_list,
    permission=permission_document_file_view, text=_(message='Files'),
    view='documents:document_file_list',
)
link_document_file_preview = Link(
    args='resolved_object.pk',
    icon=icon_document_file_preview,
    permission=permission_document_file_view,
    text=_(message='Preview'), view='documents:document_file_preview'
)
link_document_file_transformations_clear = Link(
    args='resolved_object.id',
    icon=icon_document_file_transformation_list_clear,
    permission=permission_transformation_delete,
    text=_(message='Clear transformations'),
    view='documents:document_file_transformations_clear'
)
link_document_file_multiple_transformations_clear = Link(
    icon=icon_document_file_transformation_list_clear,
    permission=permission_transformation_delete,
    text=_(message='Clear transformations'),
    view='documents:document_file_multiple_transformations_clear'
)
link_document_file_transformations_clone = Link(
    args='resolved_object.id',
    icon=icon_document_file_transformation_list_clone,
    permission=permission_transformation_edit,
    text=_(message='Clone transformations'),
    view='documents:document_file_transformations_clone'
)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .literals import DJANGO_SQLITE_BACKEND


def check_for_sqlite():
    return settings.DATABASES['default']['ENGINE'] == DJANGO_SQLITE_BACKEND and settings.DEBUG is False


def check_queryset(view, queryset):
    """
    Validate that a view queryset is usable.
    """
    try:
        queryset.query
    except AttributeError:
        # Check if it is an iterable.
        try:
            iter(queryset)
        except TypeError as exception:
            raise ImproperlyConfigured(
                'Queryset `{}` of view `{}` is not a valid queryset.'.format(
                    queryset, view.__class__
                )
            ) from exception
        else:
            return queryset
    else:
        return queryset


def instance_list_to_queryset(instance_list):
    manager = instance_list[0]._meta.default_manager

    return manager.filter(
        pk__in=[instance.pk for instance in instance_list]
    )



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Messaging'), name='messaging'
)

event_message_created = namespace.add_event_type(
    label=_(message='Message created'), name='message_created'
)
event_message_edited = namespace.add_event_type(
    label=_(message='Message edited'), name='message_edited'
)



from mayan.apps.document_indexing.models.index_instance_models import IndexInstanceNode
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin
from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .literals import (
    TEST_METADATA_INDEX_NODE_TEMPLATE, TEST_METADATA_TYPE_LABEL_EDITED,
    TEST_METADATA_VALUE, TEST_METADATA_VALUE_EDITED
)
from .mixins.document_metadata_mixins import DocumentMetadataMixin


class MetadataIndexingTestCase(
    IndexTemplateTestMixin, DocumentMetadataMixin, GenericDocumentTestCase
):
    _test_index_template_node_expression = TEST_METADATA_INDEX_NODE_TEMPLATE
    auto_create_test_document_stub = True
    auto_create_test_metadata_type = True
    auto_upload_test_document = False

    def test_indexing_no_metadata(self):
        value = '{}-{}'.format(
            self._test_metadata_type.label, TEST_METADATA_VALUE
        )

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document, value=value
            ).exists()
        )

    def test_indexing_metadata_add(self):
        self._create_test_document_metadata()
        value = '{}-{}'.format(
            self._test_metadata_type.label,
            self._test_document_metadata.value
        )

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document, value=value
            ).exists()
        )

    def test_indexing_metadata_edit(self):
        self._create_test_document_metadata()
        value = '{}-{}'.format(
            self._test_metadata_type.label,
            self._test_document_metadata.value
        )

        self._test_document_metadata.value = TEST_METADATA_VALUE_EDITED
        self._test_document_metadata.save()
        value_edited = '{}-{}'.format(
            self._test_metadata_type.label,
            self._test_document_metadata.value
        )

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document, value=value
            ).exists()
        )
        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document, value=value_edited
            ).exists()
        )

    def test_indexing_metadata_remove(self):
        self._create_test_document_metadata()
        value = '{}-{}'.format(
            self._test_metadata_type.label,
            self._test_document_metadata.value
        )

        self._test_document_metadata.delete()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document, value=value
            ).exists()
        )

    def test_indexing_metadata_type_delete(self):
        self._create_test_document_metadata()
        value = '{}-{}'.format(
            self._test_metadata_type.label,
            self._test_document_metadata.value
        )

        self._test_metadata_type.delete()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document, value=value
            ).exists()
        )

    def test_indexing_metadata_type_edit(self):
        self._create_test_document_metadata()
        value = '{}-{}'.format(
            self._test_metadata_type.label,
            self._test_document_metadata.value
        )

        self._test_metadata_type.label = TEST_METADATA_TYPE_LABEL_EDITED
        self._test_metadata_type.save()
        value_edited = '{}-{}'.format(
            self._test_metadata_type.label,
            self._test_document_metadata.value
        )

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document, value=value
            ).exists()
        )
        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document, value=value_edited
            ).exists()
        )



from django.apps import apps
from django.db.utils import OperationalError, ProgrammingError

from .literals import STORAGE_NAME_WORKFLOW_CACHE


def callback_update_workflow_image_cache_size(setting):
    Cache = apps.get_model(app_label='file_caching', model_name='Cache')

    try:
        cache = Cache.objects.get(
            defined_storage_name=STORAGE_NAME_WORKFLOW_CACHE
        )
    except (OperationalError, ProgrammingError):
        """
        Non fatal. Non initialized installation. Ignore exception.
        """
    else:
        cache.maximum_size = setting.value
        cache.save()



from mayan.apps.sources.source_backends.base import SourceBackend

from ..source_backends.storage_backend_source_mixins import SourceBackendMixinStoredFileLocationStorageBackend
from ..source_backends.stored_file_source_mixins import SourceBackendMixinStoredFileSourceBase


class SourceBackendTestStorage(
    SourceBackendMixinStoredFileLocationStorageBackend,
    SourceBackendMixinStoredFileSourceBase, SourceBackend
):
    label = 'Test storage source backend'



import hashlib
import logging

from PIL import Image, ImageColor, ImageFilter

from django import forms
from django.utils.encoding import force_bytes
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.forms import Form
from mayan.apps.views.http import URL
from mayan.apps.views.widgets import ColorWidget

from .layers import layer_decorations, layer_saved_transformations
from .transformation_mixins import (
    AssetTransformationMixin,
    ImagePasteCoordinatesAbsoluteTransformationMixin,
    ImagePasteCoordinatesPercentTransformationMixin,
    ImageWatermarkPercentTransformationMixin,
    TransformationDrawRectangleMixin
)

logger = logging.getLogger(name=__name__)


class BaseTransformationType(type):
    def __str__(self):
        return str(self.label)


class BaseTransformation(metaclass=BaseTransformationType):
    """
    Transformation can modify the appearance of the document's page preview.
    Some transformation available are: Rotate, zoom, resize and crop.
    """
    _layer_transformations = {}
    _registry = {}
    arguments = ()
    name = 'base_transformation'
    template_name = None

    @staticmethod
    def combine(transformations):
        result = hashlib.sha256()

        for transformation in transformations or ():
            try:
                result.update(
                    transformation.cache_hash()
                )
            except Exception as exception:
                logger.error(
                    'Unable to compute hash for transformation: %s; %s',
                    transformation, exception
                )

        return result.hexdigest()

    @staticmethod
    def list_as_query_string(transformation_instance_list):
        result = URL()

        for index, transformation in enumerate(transformation_instance_list):
            result.args[
                'transformation_{}_name'.format(index)
            ] = transformation.name

            for argument in transformation.arguments:
                value = getattr(transformation, argument)
                result.args[
                    'transformation_{}_argument__{}'.format(index, argument)
                ] = value

        return result.to_string()

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def get_arguments(cls):
        return cls.arguments

    @classmethod
    def get_assigned_layer(cls):
        for layer, transformations in cls._layer_transformations.items():
            if cls in transformations:
                return layer

    @classmethod
    def get_form_class(cls):
        return getattr(cls, 'Form', None)

    @classmethod
    def get_label(cls):
        arguments = cls.get_arguments()
        if arguments:
            return format_lazy(
                '{}: {}', cls.label, ', '.join(arguments)
            )
        else:
            return cls.label

    @classmethod
    def get_template_name(cls):
        return cls.template_name

    @classmethod
    def get_transformation_choices(cls, group_by_layer=False, layer=None):
        if layer:
            transformation_list = [
                (
                    transformation.name, transformation
                ) for transformation in cls._layer_transformations.get(
                    layer, ()
                )
            ]
        else:
            transformation_list = cls._registry.items()

        if group_by_layer:
            flat_transformation_list = [
                klass for name, klass in transformation_list
            ]

            layer_transformation_choices = {}
            for layer, transformations in cls._layer_transformations.items():
                for transformation in transformations:
                    if transformation in flat_transformation_list:
                        layer_transformation_choices.setdefault(
                            layer, []
                        )
                        layer_transformation_choices[layer].append(
                            (
                                transformation.name,
                                transformation.get_label()
                            )
                        )

                # Sort the transformation for each layer group.
                layer_transformation_choices[layer].sort(
                    key=lambda x: x[1]
                )

            result = [
                (
                    layer.label, transformations
                ) for layer, transformations in layer_transformation_choices.items()
            ]

            # Finally sort by transformation layer group.
            return sorted(
                result, key=lambda x: x[0]
            )
        else:
            return sorted(
                (
                    name, klass.get_label()
                ) for name, klass in transformation_list
            )

    @classmethod
    def register(cls, layer, transformation):
        cls._registry[transformation.name] = transformation
        cls._layer_transformations.setdefault(
            layer, set()
        )
        cls._layer_transformations[layer].add(transformation)

    def __init__(self, **kwargs):
        self.kwargs = {}
        for argument_name in self.__class__.get_arguments():
            setattr(
                self, argument_name, kwargs.get(argument_name)
            )
            self.kwargs[argument_name] = kwargs.get(argument_name)

    def _update_hash(self):
        result = hashlib.sha256(
            string=force_bytes(s=self.name)
        )

        # Sort arguments for guaranteed repeatability.
        for key, value in sorted(self.kwargs.items()):
            result.update(
                force_bytes(s=key)
            )
            result.update(
                force_bytes(s=value)
            )

        return result

    def cache_hash(self):
        return force_bytes(
            s=self._update_hash().hexdigest()
        )

    def execute_on(self, image):
        self.image = image
        self.aspect = 1.0 * image.size[0] / image.size[1]


class TransformationAssetPaste(
    ImagePasteCoordinatesAbsoluteTransformationMixin,
    AssetTransformationMixin, BaseTransformation
):
    label = _(message='Paste an asset (absolute coordinates)')
    name = 'paste_asset'


class TransformationAssetPastePercent(
    ImagePasteCoordinatesPercentTransformationMixin,
    AssetTransformationMixin, BaseTransformation
):
    label = _(message='Paste an asset (percent coordinates)')
    name = 'paste_asset_percent'


class TransformationAssetWatermark(
    ImageWatermarkPercentTransformationMixin, AssetTransformationMixin,
    BaseTransformation
):
    label = _(message='Paste an asset as watermark')
    name = 'paste_asset_watermark'


class TransformationCrop(BaseTransformation):
    arguments = ('left', 'top', 'right', 'bottom',)
    label = _(message='Crop')
    name = 'crop'

    class Form(Form):
        left = forms.IntegerField(
            help_text=_(message='Number of pixels to remove from the left.'),
            label=_(message='Left'), required=False
        )
        top = forms.IntegerField(
            help_text=_(message='Number of pixels to remove from the top.'),
            label=_(message='Top'), required=False
        )
        right = forms.IntegerField(
            help_text=_(
                message='Number of pixels to remove from the right.'
            ), label=_(message='Right'), required=False
        )
        bottom = forms.IntegerField(
            help_text=_(
                message='Number of pixels to remove from the bottom.'
            ), label=_(message='Bottom'), required=False
        )

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        try:
            left = int(self.left or '0')
        except ValueError:
            left = 0

        try:
            top = int(self.top or '0')
        except ValueError:
            top = 0

        try:
            right = int(self.right or '0')
        except ValueError:
            right = 0

        try:
            bottom = int(self.bottom or '0')
        except ValueError:
            bottom = 0

        if left < 0:
            left = 0

        if left > self.image.size[0] - 1:
            left = self.image.size[0] - 1

        if top < 0:
            top = 0

        if top > self.image.size[1] - 1:
            top = self.image.size[1] - 1

        if right < 0:
            right = 0

        if right > self.image.size[0] - 1:
            right = self.image.size[0] - 1

        if bottom < 0:
            bottom = 0

        if bottom > self.image.size[1] - 1:
            bottom = self.image.size[1] - 1

        # Invert right value.
        # Pillow uses left, top, right, bottom to define a viewport
        # of real coordinates.
        # We invert the right and bottom to define a viewport
        # that can crop from the right and bottom borders without
        # having to know the real dimensions of an image.
        right = self.image.size[0] - right
        bottom = self.image.size[1] - bottom

        if left > right:
            left = right - 1

        if top > bottom:
            top = bottom - 1

        logger.debug(
            'left: %f, top: %f, right: %f, bottom: %f', left, top, right,
            bottom
        )

        return self.image.crop(
            box=(left, top, right, bottom)
        )


class TransformationDrawRectangle(
    TransformationDrawRectangleMixin, BaseTransformation
):
    arguments = (
        'left', 'top', 'right', 'bottom', 'fillcolor', 'fill_transparency',
        'outlinecolor', 'outlinewidth'
    )
    label = _(message='Draw rectangle')
    name = 'draw_rectangle'

    class Form(TransformationDrawRectangleMixin.Form):
        left = forms.IntegerField(
            help_text=_(message='Left side location in pixels.'),
            label=_(message='Left'), required=False
        )
        top = forms.IntegerField(
            help_text=_(message='Top side location in pixels.'),
            label=_(message='Top'), required=False
        )
        right = forms.IntegerField(
            help_text=_(message='Right side location in pixels.'),
            label=_(message='Right'), required=False
        )
        bottom = forms.IntegerField(
            help_text=_(message='Bottom side location in pixels.'),
            label=_(message='Bottom'), required=False
        )

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        try:
            left = int(self.left or '0')
        except ValueError:
            left = 0

        try:
            top = int(self.top or '0')
        except ValueError:
            top = 0

        try:
            right = int(self.right or '0')
        except ValueError:
            right = 0

        try:
            bottom = int(self.bottom or '0')
        except ValueError:
            bottom = 0

        if left < 0:
            left = 0

        if left > self.image.size[0] - 1:
            left = self.image.size[0] - 1

        if top < 0:
            top = 0

        if top > self.image.size[1] - 1:
            top = self.image.size[1] - 1

        if right < 0:
            right = 0

        if right > self.image.size[0] - 1:
            right = self.image.size[0] - 1

        if bottom < 0:
            bottom = 0

        if bottom > self.image.size[1] - 1:
            bottom = self.image.size[1] - 1

        # Invert right value.
        # Pillow uses left, top,right, bottom to define a viewport
        # of real coordinates.
        # We invert the right and bottom to define a viewport
        # that can crop from the right and bottom borders without
        # having to know the real dimensions of an image.
        right = self.image.size[0] - right
        bottom = self.image.size[1] - bottom

        if left > right:
            left = right - 1

        if top > bottom:
            top = bottom - 1

        self.bottom = bottom
        self.left = left
        self.top = top
        self.right = right

        return super()._execute_on(self, *args, **kwargs)


class TransformationDrawRectanglePercent(
    TransformationDrawRectangleMixin, BaseTransformation
):
    arguments = (
        'left', 'top', 'right', 'bottom', 'fillcolor', 'fill_transparency',
        'outlinecolor', 'outlinewidth'
    )
    label = _(message='Draw rectangle (percent coordinates)')
    name = 'draw_rectangle_percent'

    class Form(TransformationDrawRectangleMixin.Form):
        left = forms.FloatField(
            help_text=_(message='Left side location in percent.'),
            label=_(message='Left'), required=False
        )
        top = forms.FloatField(
            help_text=_(message='Top side location in percent.'),
            label=_(message='Top'), required=False
        )
        right = forms.FloatField(
            help_text=_(message='Right side location in percent.'),
            label=_(message='Right'), required=False
        )
        bottom = forms.FloatField(
            help_text=_(message='Bottom side location in percent.'),
            label=_(message='Bottom'), required=False
        )

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        try:
            left = float(self.left or '0')
        except ValueError:
            left = 0

        try:
            top = float(self.top or '0')
        except ValueError:
            top = 0

        try:
            right = float(self.right or '0')
        except ValueError:
            right = 0

        try:
            bottom = float(self.bottom or '0')
        except ValueError:
            bottom = 0

        if left < 0:
            left = 0

        if left > 100:
            left = 100

        if top < 0:
            top = 0

        if top > 100:
            top = 100

        if right < 0:
            right = 0

        if right > 100:
            right = 100

        if bottom < 0:
            bottom = 0

        if bottom > 100:
            bottom = 100

        logger.debug(
            'left: %f, top: %f, right: %f, bottom: %f', left, top, right,
            bottom
        )

        left = left / 100.0 * self.image.size[0]
        top = top / 100.0 * self.image.size[1]

        # Invert right value.
        # Pillow uses left, top, right, bottom to define a viewport
        # of real coordinates.
        # We invert the right and bottom to define a viewport
        # that can crop from the right and bottom borders without
        # having to know the real dimensions of an image.

        right = self.image.size[0] - (
            right / 100.0 * self.image.size[0]
        )
        bottom = self.image.size[1] - (
            bottom / 100.0 * self.image.size[1]
        )

        self.bottom = bottom
        self.left = left
        self.top = top
        self.right = right

        return super()._execute_on(self, *args, **kwargs)


class TransformationFlip(BaseTransformation):
    arguments = ()
    label = _(message='Flip')
    name = 'flip'

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        return self.image.transpose(method=Image.FLIP_TOP_BOTTOM)


class TransformationGaussianBlur(BaseTransformation):
    arguments = ('radius',)
    label = _(message='Gaussian blur')
    name = 'gaussianblur'

    class Form(Form):
        radius = forms.IntegerField(
            initial=2, label=_(message='Radius'), required=True
        )

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        return self.image.filter(
            filter=ImageFilter.GaussianBlur(radius=self.radius)
        )


class TransformationLineArt(BaseTransformation):
    label = _(message='Line art')
    name = 'lineart'

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        def lut(x):
            return 0 if x < 128 else 255

        return self.image.convert(mode='L').point(lut=lut, mode='1')


class TransformationMirror(BaseTransformation):
    arguments = ()
    label = _(message='Mirror')
    name = 'mirror'

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        return self.image.transpose(method=Image.FLIP_LEFT_RIGHT)


class TransformationResize(BaseTransformation):
    arguments = ('width', 'height')
    label = _(message='Resize')
    name = 'resize'

    class Form(Form):
        width = forms.IntegerField(
            help_text=_(
                'New width in pixels.'
            ), label=_(message='Width'), required=True
        )
        height = forms.IntegerField(
            help_text=_(
                'New height in pixels.'
            ), label=_(message='Height'), required=False
        )

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        width = int(self.width)
        height = int(
            self.height or (1.0 * width / self.aspect)
        )

        factor = 1
        while self.image.size[0] / factor > 2 * width and self.image.size[1] * 2 / factor > 2 * height:
            factor *= 2

        if factor > 1:
            self.image.thumbnail(
                size=(
                    self.image.size[0] / factor,
                    self.image.size[1] / factor
                ), resample=Image.Resampling.NEAREST
            )

        self.image.thumbnail(
            size=(width, height), resample=Image.Resampling.LANCZOS
        )

        return self.image


class TransformationRotate(BaseTransformation):
    arguments = ('degrees', 'fillcolor')
    label = _(message='Rotate')
    name = 'rotate'

    class Form(Form):
        degrees = forms.IntegerField(
            help_text=_(
                'Number of degrees to rotate the image counter clockwise '
                'around its center.'
            ), label=_(message='Degrees'), required=True
        )
        fillcolor = forms.CharField(
            help_text=_(
                'Color to be used for area outside of the rotated image.'
            ), label=_(message='Fill color'), required=False,
            widget=ColorWidget()
        )

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        self.degrees = float(self.degrees or '0')
        self.degrees %= 360

        if self.degrees == 0:
            return self.image

        fillcolor_value = getattr(self, 'fillcolor', None)
        if fillcolor_value == 'None':
            fillcolor_value = None

        if fillcolor_value:
            fillcolor = ImageColor.getrgb(color=fillcolor_value)
        else:
            fillcolor = None

        return self.image.rotate(
            angle=360 - self.degrees, expand=True, fillcolor=fillcolor,
            resample=Image.Resampling.BICUBIC
        )


class TransformationRotate90(TransformationRotate):
    arguments = ()
    degrees = 90
    label = _(message='Rotate 90 degrees')
    name = 'rotate90'

    def __init__(self, **kwargs):
        super().__init__()
        self.kwargs['degrees'] = 90


class TransformationRotate180(TransformationRotate):
    arguments = ()
    degrees = 180
    label = _(message='Rotate 180 degrees')
    name = 'rotate180'

    def __init__(self, **kwargs):
        super().__init__()
        self.kwargs['degrees'] = 180


class TransformationRotate270(TransformationRotate):
    arguments = ()
    degrees = 270
    label = _(message='Rotate 270 degrees')
    name = 'rotate270'

    def __init__(self, **kwargs):
        super().__init__()
        self.kwargs['degrees'] = 270


class TransformationUnsharpMask(BaseTransformation):
    arguments = ('radius', 'percent', 'threshold')
    label = _(message='Unsharp masking')
    name = 'unsharpmask'

    class Form(Form):
        radius = forms.IntegerField(
            initial=2, help_text=_(message='The blur radius in pixels.'),
            label=_(message='Radius'), required=True
        )
        percent = forms.FloatField(
            initial=150, help_text=_(message='Unsharp strength in percent.'),
            label=_(message='Percent'), required=True
        )
        threshold = forms.IntegerField(
            initial=3, help_text=_(
                'Minimum brightness change that will be sharpened.'
            ), label=_(message='Tthreshold'), required=True
        )

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        return self.image.filter(
            ImageFilter.UnsharpMask(
                radius=self.radius, percent=self.percent,
                threshold=self.threshold
            )
        )


class TransformationZoom(BaseTransformation):
    arguments = ('percent',)
    label = _(message='Zoom')
    name = 'zoom'

    class Form(Form):
        percent = forms.FloatField(
            help_text=_(message='Zoom level in percent.'),
            label=_(message='Percent'), required=True
        )

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)

        percent = float(self.percent or '100')

        if percent == 100:
            return self.image

        decimal_value = percent / 100

        width = int(self.image.size[0] * decimal_value)
        if width < 1:
            width = 1

        height = int(self.image.size[1] * decimal_value)
        if height < 1:
            height = 1

        return self.image.resize(
            size=(width, height), resample=Image.Resampling.LANCZOS
        )


BaseTransformation.register(
    layer=layer_decorations, transformation=TransformationAssetPaste
)
BaseTransformation.register(
    layer=layer_decorations, transformation=TransformationAssetPastePercent
)
BaseTransformation.register(
    layer=layer_decorations, transformation=TransformationAssetWatermark
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationCrop
)
BaseTransformation.register(
    layer=layer_decorations,
    transformation=TransformationDrawRectangle
)
BaseTransformation.register(
    layer=layer_decorations,
    transformation=TransformationDrawRectanglePercent
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationFlip
)
BaseTransformation.register(
    layer=layer_saved_transformations,
    transformation=TransformationGaussianBlur
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationLineArt
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationMirror
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationResize
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationRotate
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationRotate90
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationRotate180
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationRotate270
)
BaseTransformation.register(
    layer=layer_saved_transformations,
    transformation=TransformationUnsharpMask
)
BaseTransformation.register(
    layer=layer_saved_transformations, transformation=TransformationZoom
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Logging'), name='logging'
)

permission_error_log_entry_delete = namespace.add_permission(
    label=_(message='Delete error log'), name='error_log_delete'
)
permission_error_log_entry_view = namespace.add_permission(
    label=_(message='View error log'), name='error_log_view'
)



from django.contrib.auth.models import Group
from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.user_management.querysets import get_user_queryset

from .classes import QuotaBackend
from .permissions import permission_quota_edit


class DocumentTypesQuotaMixin:
    @classmethod
    def get_fields(cls):
        cls.fields.update(
            {
                'document_type_all': {
                    'label': _(message='All document types'),
                    'class': 'django.forms.BooleanField', 'default': False,
                    'help_text': _(
                        'Apply to all document types. Enabling this, '
                        'the quota will ignore the individual document type'
                        ' selection.'
                    ), 'required': False,
                },
                'document_type_ids': {
                    'label': _(message='Document types'),
                    'class': 'mayan.apps.views.fields.ModelFormFieldFilteredModelMultipleChoice',
                    'kwargs': {
                        'permission': permission_quota_edit,
                        'required': False,
                        'source_queryset': DocumentType.objects.all()
                    }, 'help_text': _(
                        'Document types to which the quota will be applied.'
                    )
                }
            }
        )

        cls.field_order = list(cls.field_order)
        cls.field_order.insert(0, 'document_type_ids')
        cls.field_order.insert(0, 'document_type_all')

        return super().get_fields()

    @classmethod
    def get_widgets(cls):
        cls.widgets.update(
            {
                'document_type_ids': {
                    'class': 'django.forms.widgets.SelectMultiple',
                    'kwargs': {
                        'attrs': {'class': 'select2', 'size': 10}
                    }
                }
            }
        )

        return super().get_widgets()

    def _get_document_types(self):
        return DocumentType.objects.filter(pk__in=self.document_type_ids)

    def get_filter_text(self):
        result = super().get_filter_text()

        if self.document_type_all:
            document_type_filter_text = _(message='all document types')
        else:
            document_type_filter_text = _(
                'document types: %(document_types)s'
            ) % {
                'document_types': QuotaBackend._queryset_to_text_list(
                    queryset=self._get_document_types()
                )
            }

        result['document_type_filter_text'] = document_type_filter_text
        return result


class GroupsUsersQuotaMixin:
    @staticmethod
    def _get_user_full_name(user):
        try:
            return user.get_full_name() or user
        except AttributeError:
            return user

    @classmethod
    def get_fields(cls):
        cls.fields.update(
            {
                'group_ids': {
                    'label': _(message='Groups'),
                    'class': 'mayan.apps.views.fields.ModelFormFieldFilteredModelMultipleChoice',
                    'kwargs': {
                        'permission': permission_quota_edit,
                        'required': False,
                        'source_model': Group
                    }, 'help_text': _(
                        'Groups to which the quota will be applied.'
                    )
                },
                'user_all': {
                    'label': _(message='All users'),
                    'class': 'django.forms.BooleanField', 'default': False,
                    'help_text': _(
                        'Apply the quota to all users in the system, '
                        'excluding admins and staff. '
                        'Enabling this option, the quota will ignore the '
                        'individual user and group selection.'
                    ), 'required': False
                },
                'user_ids': {
                    'label': _(message='Users'),
                    'class': 'mayan.apps.views.fields.ModelFormFieldFilteredModelMultipleChoice',
                    'kwargs': {
                        'permission': permission_quota_edit,
                        'required': False,
                        'source_queryset': get_user_queryset()
                    }, 'help_text': _(
                        'Users to which the quota will be applied.'
                    )
                }
            }
        )

        cls.field_order = list(cls.field_order)
        cls.field_order.insert(0, 'user_ids')
        cls.field_order.insert(0, 'group_ids')
        cls.field_order.insert(0, 'user_all')

        return super().get_fields()

    @classmethod
    def get_widgets(cls):
        cls.widgets.update(
            {
                'group_ids': {
                    'class': 'django.forms.widgets.SelectMultiple',
                    'kwargs': {
                        'attrs': {'class': 'select2', 'size': 10}
                    }
                },
                'user_ids': {
                    'class': 'django.forms.widgets.SelectMultiple',
                    'kwargs': {
                        'attrs': {'class': 'select2', 'size': 10}
                    }
                }
            }
        )

        return super().get_widgets()

    def get_filter_text(self):
        result = super().get_filter_text()

        if self.user_all:
            user_filter_text = _(message='all users')
        else:
            user_filter_text = _(message='groups: %(groups)s, users: %(users)s') % {
                'groups': QuotaBackend._queryset_to_text_list(
                    queryset=self._get_groups()
                ),
                'users': QuotaBackend._queryset_to_text_list(
                    queryset=self._get_users()
                )
            }

        result['user_filter_text'] = user_filter_text
        return result

    def _get_groups(self):
        return Group.objects.filter(pk__in=self.group_ids)

    def _get_users(self):
        queryset = get_user_queryset()
        return queryset.filter(pk__in=self.user_ids)



from django.utils.safestring import mark_safe


def document_link(document):
    return mark_safe(
        s='<a href="{}">{}</a>'.format(
            document.get_absolute_url(), document
        )
    )



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    name='document_states', label=_(message='Document workflows')
)

permission_workflow_template_create = namespace.add_permission(
    name='workflow_create', label=_(message='Create workflow templates')
)
permission_workflow_template_delete = namespace.add_permission(
    name='workflow_delte', label=_(message='Delete workflow templates')
)
permission_workflow_template_edit = namespace.add_permission(
    name='workflow_edit', label=_(message='Edit workflow templates')
)
permission_workflow_template_view = namespace.add_permission(
    name='workflow_view', label=_(message='View workflow templates')
)
# Translators: This text refers to the permission to grant user the ability to
# 'transition workflows' from one state to another, to move the workflow
# forwards
permission_workflow_instance_transition = namespace.add_permission(
    name='workflow_transition', label=_(message='Transition workflow instances')
)
permission_workflow_tools = namespace.add_permission(
    name='workflow_tools', label=_(message='Execute workflow tools')
)



from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import TEST_FILE_SMALL_PATH
from mayan.apps.templating.tests.mixins import TemplateTestMixin

from .literals import TEST_SOURCE_BACKEND_PATH_DOCUMENT_BASIC
from .mixins.base_mixins import SourceTestMixin


class SourceTemplatingTestCase(
    SourceTestMixin, TemplateTestMixin, GenericDocumentTestCase
):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_DOCUMENT_BASIC
    auto_upload_test_document = False

    def test_document_property_helper(self):
        self._clear_events()

        with open(file=TEST_FILE_SMALL_PATH, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_upload',
                extra_data={
                    'document_type': self._test_document_type,
                    'file_object': file_object
                }
            )

        test_document = Document.objects.first()

        result = self._render_test_template(
            context={'document': test_document},
            template_string='{{ document.source_metadata_value_of.source_id }}'
        )

        self.assertEqual(
            int(result), self._test_source.pk
        )

    def test_document_file_property_helper(self):
        self._clear_events()

        with open(file=TEST_FILE_SMALL_PATH, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_upload',
                extra_data={
                    'document_type': self._test_document_type,
                    'file_object': file_object
                }
            )

        test_document = Document.objects.first()

        result = self._render_test_template(
            context={'document': test_document},
            template_string='{{ document.file_latest.source_metadata_value_of.source_id }}'
        )

        self.assertEqual(
            int(result), self._test_source.pk
        )



import hashlib

from django.conf import settings
from django.core import serializers
from django.utils.translation import gettext_lazy as _

from mayan.apps.templating.classes import Template


class WorkflowStateActionBusinessLogicMixin:
    def get_hash(self):
        return hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        ).hexdigest()

    def evaluate_condition(self, workflow_instance):
        if self.has_condition():
            template = Template(template_string=self.condition)
            context = {'workflow_instance': workflow_instance}
            return template.render(context=context).strip()
        else:
            return True

    def execute(self, context, workflow_instance):
        if self.evaluate_condition(workflow_instance=workflow_instance):
            context.update(
                {
                    'workflow_instance': workflow_instance
                }
            )

            try:
                backend_instance = self.get_backend_instance()
                backend_instance.execute(context=context)
            except Exception as exception:
                self.error_log.create(
                    text='{}; {}'.format(
                        exception.__class__.__name__, exception
                    )
                )

                if settings.DEBUG or settings.TESTING:
                    raise
            else:
                queryset_error_log = self.error_log.all()
                queryset_error_log.delete()

    def has_condition(self):
        return self.condition.strip()

    has_condition.help_text = _(
        'The state action will be executed, depending on the condition '
        'return value.'
    )
    has_condition.short_description = _(message='Has a condition?')



from django.db import migrations


def code_document_version_active_populate(apps, schema_editor):
    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )

    for document in Document.objects.all():
        document.version_active = document.versions.filter(active=True).only('id').first()
        document.save(
            update_fields=('version_active',)
        )


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0087_document_version_active')
    ]

    operations = [
        migrations.RunPython(
            code=code_document_version_active_populate,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0014_auto_20170206_0722')
    ]

    operations = [
        migrations.AlterField(
            model_name='sanescanner',
            name='source',
            field=models.CharField(
                blank=True, choices=[
                    ('flatbed', 'Flatbed'),
                    ('Automatic Document Feeder', 'Document feeder')
                ], default='flatbed', help_text='Selects the scan source '
                '(such as a document-feeder). If this option is not '
                'supported by your scanner, leave it blank.', max_length=32,
                verbose_name='Paper source'
            )
        )
    ]



import shutil

from django.utils.encoding import force_str
from django.utils.module_loading import import_string

from mayan.apps.converter.classes import ConverterBase
from mayan.apps.storage.utils import TemporaryFile

from .settings import setting_ocr_backend, setting_ocr_backend_arguments


class OCRBackendBase:
    @staticmethod
    def get_instance():
        return import_string(
            dotted_path=setting_ocr_backend.value
        )(**setting_ocr_backend_arguments.value)

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def execute(self, file_object, language=None, transformations=None):
        self.language = language

        if not transformations:
            transformations = []

        self.converter = ConverterBase.get_converter_class()(
            file_object=file_object
        )

        for transformation in transformations:
            self.converter.transform(transformation=transformation)

        image = self.converter.get_page()

        with TemporaryFile() as temporary_image_file:
            shutil.copyfileobj(fsrc=image, fdst=temporary_image_file)
            temporary_image_file.seek(0)

            return force_str(
                s=self._execute(image_file_object=temporary_image_file)
            )



import logging
from pathlib import Path

from django.core.files.storage import FileSystemStorage
from django.utils.translation import gettext_lazy as _

from mayan.apps.sources.exceptions import SourceException

from ..classes import SourceStoredFile

logger = logging.getLogger(name=__name__)


class SourceBackendMixinStoredFileLocationFilesystem:
    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'folder_path': {
                    'class': 'django.forms.CharField',
                    'default': '',
                    'help_text': _(
                        'Server side filesystem path.'
                    ),
                    'kwargs': {
                        'max_length': 255,
                    },
                    'label': _(message='Folder path'),
                    'required': True
                },
                'include_subdirectories': {
                    'class': 'django.forms.BooleanField',
                    'default': '',
                    'help_text': _(
                        'If checked, not only will the folder path be '
                        'scanned for files but also its subdirectories.'
                    ),
                    'label': _(message='Include subdirectories?'),
                    'required': False
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Storage'), {
                    'fields': ('folder_path', 'include_subdirectories')
                }
            ),
        )

        return fieldsets

    def get_storage_backend_arguments(self):
        return {
            'location': '{}'.format(
                self.kwargs['folder_path']
            )
        }

    def get_storage_backend_class(self):
        return FileSystemStorage

    def get_storage_backend_instance(self):
        storage_backend_arguments = self.get_storage_backend_arguments()
        storage_backend_class = self.get_storage_backend_class()

        try:
            return storage_backend_class(**storage_backend_arguments)
        except Exception as exception:
            message = _(
                'Unable to initialize storage; %s'
            ) % exception

            logger.fatal(message)
            raise TypeError(message) from exception

    def get_stored_file_list(self):
        path = Path(
            self.kwargs['folder_path']
        )

        # Force testing the path and raise errors for the log.
        path.lstat()
        if not path.is_dir():
            raise SourceException(
                'Path {} is not a directory.'.format(path)
            )

        regex_exclude = self.get_regex_exclude()
        regex_include = self.get_regex_include()

        include_subdirectories = self.kwargs.get(
            'include_subdirectories', False
        )

        try:
            if include_subdirectories:
                iterator = path.rglob(pattern='*')
            else:
                iterator = path.glob(pattern='*')

            for entry in iterator:
                if entry.is_file():
                    if regex_include.match(string=entry.name) and not regex_exclude.match(string=entry.name):
                        relative_filename = str(
                            entry.relative_to(
                                self.kwargs['folder_path']
                            )
                        )
                        yield SourceStoredFile(
                            filename=relative_filename, source=self
                        )
        except Exception as exception:
            message = 'Unable get list of files from source: {}; {}'.format(
                self, exception
            )

            logger.error(message)
            raise ValueError(message) from exception



import json

from mayan.apps.testing.tests.base import MayanMigratorTestCase

from ..source_backends import SourceBackendIMAPEmail, SourceBackendPOP3Email

from .literals import TEST_EMAIL_SOURCE_PASSWORD, TEST_EMAIL_SOURCE_USERNAME


class SourceBackendPathMigrationTestCase(MayanMigratorTestCase):
    auto_create_test_source = False
    migrate_from = ('sources', '0028_auto_20210905_0558')
    migrate_to = ('source_emails', '0004_fix_backend_paths')

    def prepare(self):
        # Manually initialize the SourceTestMixin.
        self._test_sources = []

        Source = self.old_state.apps.get_model(
            app_label='sources', model_name='Source'
        )

        Source.objects.create(
            backend_path='mayan.apps.sources.source_backends.email_backends.SourceBackendIMAPEmail',
            label='test source IMAP'
        )
        Source.objects.create(
            backend_path='mayan.apps.sources.source_backends.email_backends.SourceBackendPOP3Email',
            label='test source POP3'
        )

    def test_source_backend_path_updates(self):
        Source = self.old_state.apps.get_model(
            app_label='sources', model_name='Source'
        )

        self.assertEqual(
            Source.objects.get(label='test source IMAP').backend_path,
            SourceBackendIMAPEmail.get_class_path()
        )
        self.assertEqual(
            Source.objects.get(label='test source POP3').backend_path,
            SourceBackendPOP3Email.get_class_path()
        )


class SourceBackendCredentialMigrationTestCase(MayanMigratorTestCase):
    migrate_from = ('source_emails', '0001_update_source_backend_paths')
    migrate_to = ('source_emails', '0002_migrate_to_credentials')

    def prepare(self):
        # Manually initialize the SourceTestMixin.
        self._test_sources = []

        Source = self.old_state.apps.get_model(
            app_label='sources', model_name='Source'
        )

        Source.objects.create(
            backend_data=json.dumps(
                obj={
                    'password': TEST_EMAIL_SOURCE_PASSWORD,
                    'username': TEST_EMAIL_SOURCE_USERNAME
                }
            ),
            backend_path='mayan.apps.source_emails.tests.email_backends.SourceBackendTestEmail',
            label='test source email'
        )

    def test_source_backend_credential_migration(self):
        StoredCredential = self.old_state.apps.get_model(
            app_label='credentials', model_name='StoredCredential'
        )
        Source = self.old_state.apps.get_model(
            app_label='sources', model_name='Source'
        )

        self.assertEqual(
            StoredCredential.objects.count(), 1
        )

        self._test_source = Source.objects.first()

        source_backend_data = json.loads(s=self._test_source.backend_data)

        test_stored_credential = StoredCredential.objects.first()
        test_stored_credential_backend_data = json.loads(
            s=test_stored_credential.backend_data
        )

        self.assertEqual(
            source_backend_data['stored_credential_id'],
            test_stored_credential.pk
        )

        self.assertEqual(
            test_stored_credential_backend_data['password'],
            TEST_EMAIL_SOURCE_PASSWORD
        )
        self.assertEqual(
            test_stored_credential_backend_data['username'],
            TEST_EMAIL_SOURCE_USERNAME
        )



from django.contrib import messages
from django.db.models import Count
from django.http import HttpResponseRedirect
from django.utils.translation import gettext_lazy as _


class DocumentMetadataSameTypeViewMixin:
    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(request=request, *args, **kwargs)

        queryset = self.get_object_list()

        for document in queryset:
            document.add_as_recent_document_for_user(user=request.user)

        document_type_count = queryset.values('document_type').distinct().aggregate(
            Count('document_type')
        )['document_type__count']

        if document_type_count > 1:
            messages.error(
                message=_(
                    'Selected documents must be of the same type.'
                ), request=request
            )
            return HttpResponseRedirect(redirect_to=self.previous_url)

        return result



import uuid

from django.db import migrations, models

import mayan.apps.storage.classes


def UUID_FUNCTION(*args, **kwargs):
    return str(
        uuid.uuid4()
    )


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0047_auto_20180917_0737')
    ]

    operations = [
        migrations.AlterField(
            model_name='documentversion',
            name='file',
            field=models.FileField(
                storage=mayan.apps.storage.classes.FakeStorageSubclass(),
                upload_to=UUID_FUNCTION,
                verbose_name='File'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.dependencies.classes import (
    BinaryDependency, PythonDependency
)

from .backends.python_gnupg import gpg_path

BinaryDependency(
    label='GNU privacy guard', help_text=_(
        'GNU privacy guard - a PGP implementation.'
    ), module=__name__, name='gnupg1', path=gpg_path
)
PythonDependency(
    legal_text='''
        Copyright (c) 2008-2014 by Vinay Sajip.
        All rights reserved.

        Redistribution and use in source and binary forms, with or without
        modification, are permitted provided that the following conditions are met:

            * Redistributions of source code must retain the above copyright notice,
              this list of conditions and the following disclaimer.
            * Redistributions in binary form must reproduce the above copyright notice,
              this list of conditions and the following disclaimer in the documentation
              and/or other materials provided with the distribution.
            * The name(s) of the copyright holder(s) may not be used to endorse or
              promote products derived from this software without specific prior
              written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) "AS IS" AND ANY EXPRESS OR
        IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
        MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
        EVENT SHALL THE COPYRIGHT HOLDER(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
        INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
        LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
        PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
        LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
        OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
        ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    ''', module=__name__, name='python_gnupg', version_string='==0.4.9'
)



from django.shortcuts import get_object_or_404

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import (
    permission_document_type_view, permission_document_view
)
from mayan.apps.documents.serializers.document_type_serializers import DocumentTypeSerializer
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from .models import ResolvedSmartLink, SmartLink
from .permissions import (
    permission_resolved_smart_link_view, permission_smart_link_create,
    permission_smart_link_delete, permission_smart_link_edit,
    permission_smart_link_view
)
from .serializers import (
    ResolvedSmartLinkDocumentSerializer, ResolvedSmartLinkSerializer,
    SmartLinkConditionSerializer, SmartLinkDocumentTypeAddSerializer,
    SmartLinkDocumentTypeRemoveSerializer, SmartLinkSerializer
)


class APIDocumentResolvedSmartLinkDetailView(
    ExternalObjectAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Return the details of the selected resolved smart link.
    """
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    lookup_url_kwarg = 'resolved_smart_link_id'
    mayan_external_object_permission_map = {
        'GET': permission_resolved_smart_link_view
    }
    mayan_object_permission_map = {
        'GET': permission_resolved_smart_link_view
    }
    serializer_class = ResolvedSmartLinkSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'document': self.get_external_object()
        }

    def get_serializer_extra_context(self):
        if self.kwargs:
            return {
                'document': self.get_external_object()
            }
        else:
            return {}

    def get_source_queryset(self):
        return ResolvedSmartLink.objects.get_for(
            document=self.get_external_object()
        )


class APIDocumentResolvedSmartLinkDocumentListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of the smart link documents that apply to the document.
    """
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    mayan_external_object_permission_map = {
        'GET': permission_resolved_smart_link_view
    }
    mayan_object_permission_map = {'GET': permission_document_view}
    serializer_class = ResolvedSmartLinkDocumentSerializer

    def get_resolved_smart_link(self):
        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_resolved_smart_link_view,
            queryset=SmartLink.objects.filter(enabled=True),
            user=self.request.user
        )

        return get_object_or_404(
            klass=queryset, pk=self.kwargs['resolved_smart_link_id']
        )

    def get_serializer_extra_context(self):
        if self.kwargs:
            return {
                'document': self.get_external_object(),
                'resolved_smart_link': self.get_resolved_smart_link()
            }
        else:
            return {}

    def get_source_queryset(self):
        return self.get_resolved_smart_link().get_linked_documents_for(
            document=self.get_external_object()
        )


class APIDocumentResolvedSmartLinkListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of the smart links that apply to the document.
    """
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    lookup_url_kwarg = 'resolved_smart_link_id'
    mayan_external_object_permission_map = {
        'GET': permission_resolved_smart_link_view
    }
    mayan_object_permission_map = {
        'GET': permission_resolved_smart_link_view
    }
    serializer_class = ResolvedSmartLinkSerializer

    def get_serializer_extra_context(self):
        if self.kwargs:
            return {
                'document': self.get_external_object()
            }
        else:
            return {}

    def get_source_queryset(self):
        return ResolvedSmartLink.objects.get_for(
            document=self.get_external_object()
        )


class APISmartLinkConditionListView(
    ExternalObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the smart link conditions.
    post: Create a new smart link condition.
    """
    external_object_class = SmartLink
    external_object_pk_url_kwarg = 'smart_link_id'
    mayan_external_object_permission_map = {
        'GET': permission_smart_link_view,
        'POST': permission_smart_link_edit
    }
    ordering_fields = ('enabled', 'id')
    serializer_class = SmartLinkConditionSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'smart_link': self.get_external_object()
        }

    def get_source_queryset(self):
        return self.get_external_object().conditions.all()


class APISmartLinkConditionView(
    ExternalObjectAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected smart link condition.
    get: Return the details of the selected smart link condition.
    patch: Edit the selected smart link condition.
    put: Edit the selected smart link condition.
    """
    external_object_class = SmartLink
    external_object_pk_url_kwarg = 'smart_link_id'
    lookup_url_kwarg = 'smart_link_condition_id'
    mayan_external_object_permission_map = {
        'DELETE': permission_smart_link_edit,
        'GET': permission_smart_link_view,
        'PATCH': permission_smart_link_edit,
        'PUT': permission_smart_link_edit
    }
    serializer_class = SmartLinkConditionSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'smart_link': self.get_external_object()
        }

    def get_source_queryset(self):
        return self.get_external_object().conditions.all()


class APISmartLinkListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the smart links.
    post: Create a new smart link.
    """
    mayan_object_permission_map = {'GET': permission_smart_link_view}
    mayan_view_permission_map = {'POST': permission_smart_link_create}
    serializer_class = SmartLinkSerializer
    source_queryset = SmartLink.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APISmartLinkDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected smart link.
    get: Return the details of the selected smart link.
    patch: Edit the selected smart link.
    put: Edit the selected smart link.
    """
    lookup_url_kwarg = 'smart_link_id'
    mayan_object_permission_map = {
        'DELETE': permission_smart_link_delete,
        'GET': permission_smart_link_view,
        'PATCH': permission_smart_link_edit,
        'PUT': permission_smart_link_edit
    }
    ordering_fields = ('dynamic_label', 'enabled', 'id', 'label')
    serializer_class = SmartLinkSerializer
    source_queryset = SmartLink.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APISmartLinkDocumentTypeAddView(generics.ObjectActionAPIView):
    """
    post: Add a document type to a smart link.
    """
    lookup_url_kwarg = 'smart_link_id'
    mayan_object_permission_map = {'POST': permission_smart_link_edit}
    serializer_class = SmartLinkDocumentTypeAddSerializer
    source_queryset = SmartLink.objects.all()

    def object_action(self, obj, request, serializer):
        document_type = serializer.validated_data['document_type']
        obj.document_types_add(
            queryset=DocumentType.objects.filter(pk=document_type.pk),
            user=self.request.user
        )


class APISmartLinkDocumentTypeListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Return a list of the selected smart link document types.
    """
    external_object_class = SmartLink
    external_object_pk_url_kwarg = 'smart_link_id'
    mayan_external_object_permission_map = {
        'GET': permission_smart_link_view
    }
    mayan_object_permission_map = {'GET': permission_document_type_view}
    serializer_class = DocumentTypeSerializer

    def get_source_queryset(self):
        return self.get_external_object().document_types.all()


class APISmartLinkDocumentTypeRemoveView(generics.ObjectActionAPIView):
    """
    post: Remove a document type from a smart link.
    """
    lookup_url_kwarg = 'smart_link_id'
    mayan_object_permission_map = {'POST': permission_smart_link_edit}
    serializer_class = SmartLinkDocumentTypeRemoveSerializer
    source_queryset = SmartLink.objects.all()

    def object_action(self, obj, request, serializer):
        document_type = serializer.validated_data['document_type']
        obj.document_types_remove(
            queryset=DocumentType.objects.filter(pk=document_type.pk),
            user=self.request.user
        )



from django.db import models
from django.utils.translation import gettext_lazy as _


class UpdatedSetting(models.Model):
    global_name = models.CharField(
        db_index=True, help_text=_(
            'A short text used as the tag name.'
        ), max_length=255, unique=True, verbose_name=_(message='Global name')
    )
    value_new = models.TextField(
        blank=True, null=True, verbose_name=_(message='New value')
    )
    value_old = models.TextField(
        blank=True, null=True, verbose_name=_(message='Old value')
    )

    class Meta:
        verbose_name = _(message='Updated setting')
        verbose_name_plural = _(message='Updated settings')



from mayan.apps.appearance.classes import Icon

icon_user_theme_settings_detail = Icon(
    driver_name='fontawesome', symbol='palette'
)
icon_user_theme_settings_edit = Icon(
    driver_name='fontawesome-dual', primary_symbol='palette',
    secondary_symbol='pencil-alt'
)

icon_menu_actions = Icon(driver_name='fontawesome', symbol='ellipsis-v')
icon_menu_views = Icon(driver_name='fontawesome', symbol='eye')

icon_theme = Icon(driver_name='fontawesome', symbol='palette')
icon_theme_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='palette',
    secondary_symbol='plus'
)
icon_theme_delete = Icon(driver_name='fontawesome', symbol='times')
icon_theme_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_theme_setup = icon_theme
icon_theme_list = icon_theme



from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.events.classes import ModelEventType

from ..events import event_index_template_edited
from ..permissions import permission_index_template_edit

from .mixins import (
    IndexTemplateEventTriggerViewTestMixin, IndexTemplateTestMixin
)


class IndexTemplateEventTriggerViewTestCase(
    IndexTemplateTestMixin, IndexTemplateEventTriggerViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_index_template_event_trigger_get_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_event_trigger_get_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_event_trigger_get_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_event_trigger_get_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_event_trigger_add_view_no_permission(self):
        stored_event_type_id = ModelEventType.get_for_class(
            klass=Document
        )[3].stored_event_type.pk

        self._test_index_template.event_triggers.get(
            stored_event_type_id=stored_event_type_id
        ).delete()

        test_index_template_event_trigger_count = self._test_index_template.event_triggers.count()

        self._clear_events()

        response = self._request_test_index_template_event_trigger_add_view(
            stored_event_type_id=stored_event_type_id
        )
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_index_template.event_triggers.count(),
            test_index_template_event_trigger_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_event_trigger_add_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        stored_event_type_id = ModelEventType.get_for_class(
            klass=Document
        )[3].stored_event_type.pk

        self._test_index_template.event_triggers.get(
            stored_event_type_id=stored_event_type_id
        ).delete()

        test_index_template_event_trigger_count = self._test_index_template.event_triggers.count()

        self._clear_events()

        response = self._request_test_index_template_event_trigger_add_view(
            stored_event_type_id=stored_event_type_id
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_index_template.event_triggers.count(),
            test_index_template_event_trigger_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)

    def test_index_template_event_trigger_remove_view_no_permission(self):
        test_index_template_event_trigger_count = self._test_index_template.event_triggers.count()

        stored_event_type_id = ModelEventType.get_for_class(
            klass=Document
        )[3].stored_event_type.pk

        self._clear_events()

        response = self._request_test_index_template_event_trigger_remove_view(
            stored_event_type_id=stored_event_type_id
        )
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_index_template.event_triggers.count(),
            test_index_template_event_trigger_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_event_trigger_remove_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        test_index_template_event_trigger_count = self._test_index_template.event_triggers.count()

        stored_event_type_id = ModelEventType.get_for_class(
            klass=Document
        )[3].stored_event_type.pk

        self._clear_events()

        response = self._request_test_index_template_event_trigger_remove_view(
            stored_event_type_id=stored_event_type_id
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_index_template.event_triggers.count(),
            test_index_template_event_trigger_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)



import logging

from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.document_states.classes import WorkflowAction
from mayan.apps.document_states.literals import BASE_WORKFLOW_TEMPLATE_STATE_ACTION_HELP_TEXT
from mayan.apps.document_states.models.workflow_instance_models import WorkflowInstance
from mayan.apps.user_management.querysets import get_user_queryset

from .models import Message

logger = logging.getLogger(name=__name__)


class WorkflowActionMessageSend(WorkflowAction):
    form_fields = {
        'username_list': {
            'label': _(message='Username list'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    format_lazy(
                        '{}. {}',
                        _(
                            'Comma separated list of usernames that will '
                            'receive the message.'
                        ),
                        BASE_WORKFLOW_TEMPLATE_STATE_ACTION_HELP_TEXT
                    )
                ),
                'model': WorkflowInstance,
                'model_variable': 'workflow_instance',
                'required': True
            }
        },
        'subject': {
            'label': _(message='Subject'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    format_lazy(
                        '{}. {}',
                        _(
                            'Subject of the message to be sent.'
                        ), BASE_WORKFLOW_TEMPLATE_STATE_ACTION_HELP_TEXT
                    )
                ),
                'model': WorkflowInstance,
                'model_variable': 'workflow_instance',
                'required': True
            }
        },
        'body': {
            'label': _(message='Body'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    format_lazy(
                        '{}. {}',
                        _(
                            'The actual text to send.'
                        ), BASE_WORKFLOW_TEMPLATE_STATE_ACTION_HELP_TEXT
                    )
                ),
                'model': WorkflowInstance,
                'model_variable': 'workflow_instance',
                'required': True
            }
        }
    }
    label = _(message='Send user message')

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Recipients'), {
                    'fields': ('username_list',)
                },
            ), (
                _(message='Content'), {
                    'fields': ('subject', 'body')
                }
            )
        )
        return fieldsets

    def execute(self, context):
        username_list = self.render_field(
            field_name='username_list', context=context
        ) or ''
        username_list = username_list.split(',')

        subject = self.render_field(
            field_name='subject', context=context
        ) or ''

        body = self.render_field(
            field_name='body', context=context
        ) or ''

        for user in get_user_queryset().filter(username__in=username_list):
            Message.objects.create(
                user=user, body=body, subject=subject
            )



import logging

from django.contrib import messages
from django.core.files import File
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.literals import DEFAULT_DOCUMENT_FILE_ACTION_NAME
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_file_new
from mayan.apps.views.generics import SingleObjectListView
from mayan.apps.views.utils import request_is_ajax
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms import NewDocumentFileForm
from ..icons import (
    icon_document_file_source_metadata_list, icon_document_file_upload
)
from ..models import Source
from ..permissions import permission_sources_metadata_view

from .base import UploadBaseView

logger = logging.getLogger(name=__name__)


class DocumentFileSourceMetadataList(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_permission = permission_sources_metadata_view
    external_object_pk_url_kwarg = 'document_file_id'
    external_object_queryset = DocumentFile.valid.all()
    view_icon = icon_document_file_source_metadata_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_document_file_source_metadata_list,
            'no_results_text': _(
                'This means that the sources system did not record any '
                'information about the creation of the document file.'
            ),
            'no_results_title': _(
                'No source metadata available for this document file.'
            ),
            'object': self.external_object,
            'title': _(
                'Source metadata for: %(document_file)s'
            ) % {
                'document_file': self.external_object
            }
        }

    def get_source_queryset(self):
        queryset_document_source_metadata = self.external_object.source_metadata.all()

        queryset_sources = AccessControlList.objects.restrict_queryset(
            permission=permission_sources_metadata_view,
            queryset=Source.objects.all(),
            user=self.request.user
        )

        return queryset_document_source_metadata.filter(
            source__in=queryset_sources
        )


class DocumentFileUploadView(ExternalObjectViewMixin, UploadBaseView):
    document_form = NewDocumentFileForm
    external_object_queryset = Document.valid.all()
    external_object_permission = permission_document_file_new
    external_object_pk_url_kwarg = 'document_id'
    object_permission = permission_document_file_new
    source_link_view_name = 'sources:document_file_upload'
    view_icon = icon_document_file_upload
    view_source_action = 'document_file_upload'

    def dispatch(self, request, *args, **kwargs):
        self.subtemplates_list = []
        result = super().dispatch(request=request, *args, **kwargs)

        try:
            DocumentFile.execute_pre_create_hooks(
                kwargs={
                    'document': self.external_object,
                    'file_object': None,
                    'user': self.request.user
                }
            )
        except Exception as exception:
            messages.error(
                message=_(
                    'Unable to upload new files for document '
                    '"%(document)s". %(exception)s'
                ) % {'document': self.external_object, 'exception': exception},
                request=self.request
            )
            return HttpResponseRedirect(
                redirect_to=reverse(
                    viewname='documents:document_file_list',
                    kwargs={'document_id': self.external_object.pk}
                )
            )

        return result

    def forms_valid(self, forms):
        action = self.source.get_action(
            name='document_file_upload'
        )

        interface_load_kwargs = {
            'document': self.external_object, 'forms': forms,
            'request': self.request, 'view': self
        }

        try:
            source_form = forms.get('source_form')
            form_file_object = None
            if source_form:
                form_file_data = source_form.cleaned_data.get('file')
                if form_file_data:
                    form_file_object = File(file=form_file_data)

            DocumentFile.execute_pre_create_hooks(
                kwargs={
                    'document': self.external_object,
                    'file_object': form_file_object,
                    'user': self.request.user
                }
            )
            action.execute(
                interface_name='View',
                interface_load_kwargs=interface_load_kwargs
            )
        except Exception as exception:
            message = _(
                'Error executing document file upload task; '
                '%(exception)s'
            ) % {
                'exception': exception,
            }
            logger.critical(msg=message, exc_info=True)
            if request_is_ajax(request=self.request):
                return JsonResponse(
                    data={
                        'error': str(message)
                    }, status=500
                )
            else:
                messages.error(
                    message=message.replace('\n', ' '),
                    request=self.request
                )
                raise type(exception)(message)
        else:
            messages.success(
                message=_(
                    'New document file queued for upload and will be '
                    'available shortly.'
                ), request=self.request
            )

            return HttpResponseRedirect(
                redirect_to=reverse(
                    viewname='documents:document_file_list', kwargs={
                        'document_id': self.external_object.pk
                    }
                )
            )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                'form_action': '{}?{}'.format(
                    reverse(
                        viewname=self.request.resolver_match.view_name,
                        kwargs=self.request.resolver_match.kwargs
                    ), self.request.META['QUERY_STRING']
                ),
                'object': self.external_object,
                'title': _(
                    'Upload a new file for document "%(document)s" '
                    'from source: %(source)s'
                ) % {
                    'document': self.external_object,
                    'source': self.source.label
                },
                'submit_label': _(message='Submit')
            }
        )
        context.update(
            self.source.get_backend_instance().get_view_context(
                context=context, request=self.request
            )
        )

        return context

    def get_form_extra_kwargs__source_form(self, **kwargs):
        return {
            'source': self.source
        }

    def get_initial__document_form(self):
        return {'action': DEFAULT_DOCUMENT_FILE_ACTION_NAME}

    def get_source_link_view_kwargs(self):
        return {
            'document_id': str(self.external_object.pk)
        }



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('locales', '0003_auto_20210130_0333')
    ]

    operations = [
        migrations.AlterField(
            model_name='userlocaleprofile',
            name='user',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='locale_profile', to=settings.AUTH_USER_MODEL, verbose_name='User')
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Document parsing'), name='document_parsing'
)

event_parsing_document_file_content_deleted = namespace.add_event_type(
    label=_(message='Document file parsed content deleted'),
    name='document_content_deleted'
)
event_parsing_document_file_submitted = namespace.add_event_type(
    label=_(message='Document file submitted for parsing'), name='version_submit'
)
event_parsing_document_file_finished = namespace.add_event_type(
    label=_(message='Document file parsing finished'), name='version_finish'
)



import json
from unittest import mock

from mayan.apps.credentials.tests.mixins import StoredCredentialPasswordUsernameTestMixin
from mayan.apps.testing.tests.base import BaseTestCase, GenericViewTestCase
from mayan.apps.testing.tests.mixins import TestServerTestCaseMixin
from mayan.apps.testing.tests.mocks import request_method_factory

from ..literals import WORKFLOW_ACTION_ON_ENTRY
from ..models.workflow_instance_models import WorkflowInstance
from ..models.workflow_models import Workflow
from ..permissions import permission_workflow_template_edit
from ..workflow_actions import (
    DocumentPropertiesEditAction, DocumentWorkflowLaunchAction, HTTPAction
)

from .literals import (
    TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH,
    TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEMPLATE_DATA,
    TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_DATA,
    TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_DESCRIPTION,
    TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_LABEL,
    TEST_HEADERS_AUTHENTICATION_KEY, TEST_HEADERS_AUTHENTICATION_VALUE,
    TEST_HEADERS_JSON, TEST_HEADERS_JSON_TEMPLATE,
    TEST_HEADERS_JSON_TEMPLATE_KEY, TEST_HEADERS_KEY, TEST_HEADERS_VALUE,
    TEST_PAYLOAD_JSON, TEST_PAYLOAD_TEMPLATE_DOCUMENT_LABEL,
    TEST_SERVER_PASSWORD, TEST_SERVER_USERNAME
)
from .mixins.workflow_template_state_action_mixins import (
    WorkflowTemplateStateActionLaunchViewTestMixin,
    WorkflowTemplateStateActionTestMixin,
    WorkflowTemplateStateActionViewTestMixin
)
from .mixins.workflow_template_transition_mixins import WorkflowTemplateTransitionTestMixin


class HTTPWorkflowActionTestCase(
    TestServerTestCaseMixin, WorkflowTemplateStateActionTestMixin,
    GenericViewTestCase
):
    auto_add_test_view = True

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_simple(self, mock_object):
        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'url': self.testserver_url
            }
        )

        self.assertFalse(self.test_view_request is None)

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_payload_simple(self, mock_object):
        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'payload': TEST_PAYLOAD_JSON,
                'url': self.testserver_url
            }
        )

        self.assertEqual(
            json.loads(s=self.test_view_request.body),
            {'label': 'label'}
        )

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_payload_template(self, mock_object):
        self._create_test_document_stub()

        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'payload': TEST_PAYLOAD_TEMPLATE_DOCUMENT_LABEL,
                'url': self.testserver_url
            }
        )

        self.assertEqual(
            json.loads(s=self.test_view_request.body),
            {'label': self._test_document.label}
        )

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_headers_simple(self, mock_object):
        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'headers': TEST_HEADERS_JSON,
                'url': self.testserver_url
            }
        )

        self.assertTrue(
            TEST_HEADERS_KEY in self.test_view_request.META,
        )
        self.assertEqual(
            self.test_view_request.META[TEST_HEADERS_KEY], TEST_HEADERS_VALUE
        )

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_headers_template(self, mock_object):
        self._create_test_document_stub()

        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'headers': TEST_HEADERS_JSON_TEMPLATE,
                'url': self.testserver_url
            }
        )
        self.assertTrue(
            TEST_HEADERS_JSON_TEMPLATE_KEY in self.test_view_request.META,
        )
        self.assertEqual(
            self.test_view_request.META[TEST_HEADERS_JSON_TEMPLATE_KEY],
            self._test_document.label
        )

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_authentication(self, mock_object):
        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'password': TEST_SERVER_PASSWORD,
                'url': self.testserver_url,
                'username': TEST_SERVER_USERNAME,
            }
        )
        self.assertTrue(
            TEST_HEADERS_AUTHENTICATION_KEY in self.test_view_request.META,
        )
        self.assertEqual(
            self.test_view_request.META[TEST_HEADERS_AUTHENTICATION_KEY],
            TEST_HEADERS_AUTHENTICATION_VALUE
        )

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_timeout_value_int(self, mock_object):
        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'timeout': '1',
                'url': self.testserver_url
            }
        )
        self.assertEqual(self.timeout, 1)

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_timeout_value_float(self, mock_object):
        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'timeout': '1.5',
                'url': self.testserver_url
            }
        )
        self.assertEqual(self.timeout, 1.5)

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_timeout_value_none(self, mock_object):
        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'url': self.testserver_url
            }
        )
        self.assertEqual(self.timeout, None)


class HTTPCredentialTemplateWorkflowActionTestCase(
    StoredCredentialPasswordUsernameTestMixin, TestServerTestCaseMixin,
    WorkflowTemplateStateActionTestMixin, GenericViewTestCase
):
    auto_add_test_view = True

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_headers_template(self, mock_object):
        self._create_test_document_stub()

        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'headers': '{{"{}": "{}"}}'.format(
                    TEST_HEADERS_JSON_TEMPLATE_KEY, '{{ credential.password }}'
                ),
                'method': 'POST',
                'stored_credential_id': self._test_stored_credential.pk,
                'url': self.testserver_url
            }
        )
        self.assertTrue(
            TEST_HEADERS_JSON_TEMPLATE_KEY in self.test_view_request.META,
        )
        self.assertEqual(
            self.test_view_request.META[TEST_HEADERS_JSON_TEMPLATE_KEY],
            self._test_stored_credential._password
        )

    @mock.patch('requests.sessions.Session.get_adapter')
    def test_http_post_action_authentication(self, mock_object):
        mock_object.side_effect = request_method_factory(test_case=self)

        self._execute_workflow_template_state_action(
            klass=HTTPAction, kwargs={
                'method': 'POST',
                'password': '{{ credential.password }}',
                'stored_credential_id': self._test_stored_credential.pk,
                'url': self.testserver_url,
                'username': '{{ credential.username }}'
            }
        )
        self.assertTrue(
            TEST_HEADERS_AUTHENTICATION_KEY in self.test_view_request.META,
        )
        self.assertEqual(
            self.test_view_request.META[TEST_HEADERS_AUTHENTICATION_KEY],
            'Basic dGVzdF9jcmVkZW50aWFsX3VzZXJuYW1lOnRlc3RfY3JlZGVudGlhbF9wYXNzd29yZA=='
        )


class HTTPWorkflowActionViewTestCase(
    WorkflowTemplateStateActionViewTestMixin, GenericViewTestCase
):
    def test_http_workflow_state_action_create_post_view_no_permission(self):
        action_count = self._test_workflow_template_state.actions.count()

        response = self._request_test_workflow_template_state_action_create_post_view(
            backend_path='mayan.apps.document_states.workflow_actions.HTTPAction',
            extra_data={
                'method_template': 'POST', 'timeout_template': '0',
                'url_template': '127.0.0.1'
            }
        )
        self.assertEqual(response.status_code, 404)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count
        )

    def test_http_workflow_state_action_create_post_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )
        action_count = self._test_workflow_template_state.actions.count()

        response = self._request_test_workflow_template_state_action_create_post_view(
            backend_path='mayan.apps.document_states.workflow_actions.HTTPAction',
            extra_data={
                'method_template': 'POST', 'timeout_template': '0',
                'url_template': '127.0.0.1'
            }
        )
        self.assertEqual(response.status_code, 302)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count + 1
        )


class DocumentPropertiesEditActionTestCase(
    WorkflowTemplateStateActionTestMixin,
    WorkflowTemplateTransitionTestMixin, BaseTestCase
):
    def test_document_properties_edit_action_field_literals(self):
        test_values = self._model_instance_to_dictionary(
            instance=self._test_document
        )

        self._execute_workflow_template_state_action(
            klass=DocumentPropertiesEditAction,
            kwargs=TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_DATA
        )

        self._test_document.refresh_from_db()

        self.assertNotEqual(
            test_values, self._model_instance_to_dictionary(
                instance=self._test_document
            )
        )

    def test_document_properties_edit_action_field_templates(self):
        label = self._test_document.label

        self._execute_workflow_template_state_action(
            klass=DocumentPropertiesEditAction,
            kwargs=TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEMPLATE_DATA
        )

        self.assertEqual(
            self._test_document.label,
            '{} new'.format(label)
        )
        self.assertEqual(
            self._test_document.description,
            label
        )

    def test_document_properties_edit_action_workflow_transition(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()

        self._test_workflow_template_states[1].actions.create(
            backend_data=json.dumps(
                obj=TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_DATA
            ),
            backend_path=TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH,
            label='', when=WORKFLOW_ACTION_ON_ENTRY,
        )

        test_workflow_instance = self._test_document.workflows.first()
        test_workflow_instance.do_transition(
            transition=self._test_workflow_template_transition
        )

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.label,
            TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_LABEL
        )
        self.assertEqual(
            self._test_document.description,
            TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_DESCRIPTION
        )


class DocumentWorkflowLaunchActionTestCase(
    WorkflowTemplateStateActionTestMixin, BaseTestCase
):
    def test_document_workflow_launch_action(self):
        self._create_test_workflow_template(
            add_test_document_type=True, auto_launch=False
        )
        self._create_test_workflow_template_state()

        workflow_count = self._test_document.workflows.count()

        self._execute_workflow_template_state_action(
            klass=DocumentWorkflowLaunchAction,
            kwargs={
                'workflows': Workflow.objects.exclude(pk=self._test_workflow_instance.workflow.pk)
            }
        )

        self.assertEqual(
            self._test_document.workflows.count(), workflow_count + 1
        )


class DocumentWorkflowLaunchActionViewTestCase(
    WorkflowTemplateStateActionLaunchViewTestMixin, GenericViewTestCase
):
    def test_document_workflow_launch_action_view_with_full_access(self):
        self._create_test_workflow_template(
            add_test_document_type=True, auto_launch=False
        )

        self._create_test_workflow_template(
            add_test_document_type=True, auto_launch=False
        )
        self._create_test_workflow_template_state()

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        action_count = self._test_workflow_template_state.actions.count()

        response = self._request_document_workflow_template_launch_action_create_view(
            extra_data={
                'workflows': self._test_workflow_templates[0].pk
            }
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_workflow_template_state.actions.count(),
            action_count + 1
        )

    def test_document_workflow_launch_action_view_and_document_create_with_full_access(self):
        self._test_document.delete()  # Send test document to trash.
        self._test_document.delete()  # Delete test document.

        self._test_workflow_template.delete()  # Delete templates.
        self._test_workflow_templates = []  # Delete templates.

        self._create_test_workflow_template(
            add_test_document_type=True, auto_launch=False
        )

        self._create_test_workflow_template(
            add_test_document_type=True, auto_launch=True
        )
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        action_count = self._test_workflow_template_state.actions.count()
        workflow_instance_count = WorkflowInstance.objects.count()

        response = self._request_document_workflow_template_launch_action_create_view(
            extra_data={
                'workflows': self._test_workflow_templates[0].pk
            }
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_workflow_template_state.actions.count(),
            action_count + 1
        )

        self._create_test_document_stub()

        self.assertEqual(
            WorkflowInstance.objects.count(), workflow_instance_count + 2
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Appearance'), name='appearance'
)

permission_theme_create = namespace.add_permission(
    label=_(message='Create new themes'), name='theme_create'
)
permission_theme_delete = namespace.add_permission(
    label=_(message='Delete themes'), name='theme_delete'
)
permission_theme_edit = namespace.add_permission(
    label=_(message='Edit themes'), name='theme_edit'
)
permission_theme_view = namespace.add_permission(
    label=_(message='View existing themes'), name='theme_view'
)



from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.model_mixins import BackendModelMixin
from mayan.apps.common.validators import validate_internal_name
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from .classes import CredentialBackendNull
from .events import event_credential_created, event_credential_edited
from .model_mixins import StoredCredentialBusinessLogicMixin


class StoredCredential(
    StoredCredentialBusinessLogicMixin, BackendModelMixin,
    ExtraDataModelMixin, models.Model
):
    _backend_model_null_backend = CredentialBackendNull

    label = models.CharField(
        help_text=_(message='Short description of this credential.'), max_length=128,
        unique=True, verbose_name=_(message='Label')
    )
    internal_name = models.CharField(
        db_index=True, help_text=_(
            'This value will be used by other apps to reference this '
            'credential. Can only contain letters, numbers, and underscores.'
        ), max_length=255, unique=True, validators=[validate_internal_name],
        verbose_name=_(message='Internal name')
    )

    class Meta:
        ordering = ('label',)
        verbose_name = _(message='Credential')
        verbose_name_plural = _(message='Credentials')

    def __str__(self):
        return self.label

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_credential_created,
            'target': 'self'
        },
        edited={
            'event': event_credential_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



from mayan.apps.common.tests.mixins import PropertyHelperTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from .mixins import DocumentFileMetadataTestMixin


class DocumentPropertyHelperTestCase(
    DocumentFileMetadataTestMixin, PropertyHelperTestMixin, BaseTestCase
):
    _test_document_file_metadata_create_auto = True

    def test_basic(self):
        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document.file_metadata_value_of,
                self._test_document_file_metadata_path
            ), self._test_document_file_metadata.value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_no_file(self):
        self._test_document_file.delete()

        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document.file_metadata_value_of,
                self._test_document_file_metadata_path
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_no_source_metadata(self):
        self._test_document_file_metadata.delete()

        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document.file_metadata_value_of,
                self._test_document_file_metadata_path
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_wrong_source_metadata_name(self):
        self._test_document_file_metadata.delete()

        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document.file_metadata_value_of,
                'invalid'
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentFilePropertyHelperTestCase(
    DocumentFileMetadataTestMixin, PropertyHelperTestMixin, BaseTestCase
):
    _test_document_file_metadata_create_auto = True

    def test_basic(self):
        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document_file.file_metadata_value_of,
                self._test_document_file_metadata_path
            ), self._test_document_file_metadata.value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_no_source_metadata(self):
        self._test_document_file_metadata.delete()

        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document_file.file_metadata_value_of,
                self._test_document_file_metadata_path
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_wrong_source_metadata_name(self):
        self._test_document_file_metadata.delete()

        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document_file.file_metadata_value_of,
                'invalid'
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from ..permissions import permission_document_view

from .base import GenericDocumentViewTestCase
from .mixins.recently_accessed_document_mixins import RecentlyAccessedDocumentViewTestMixin


class RecentlyAccessedDocumentViewTestCase(
    RecentlyAccessedDocumentViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_recently_accessed_document_list_view_no_permission(self):
        self._test_document.add_as_recent_document_for_user(
            user=self._test_case_user
        )

        self._clear_events()

        response = self._request_test_recently_accessed_document_list_view()
        self.assertNotContains(
            response=response, text=self._test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_recently_accessed_document_list_view_with_access(self):
        self._test_document.add_as_recent_document_for_user(
            user=self._test_case_user
        )

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_recently_accessed_document_list_view()
        self.assertContains(
            response=response, text=self._test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_recently_accessed_document_list_view_with_access(self):
        self._test_document.add_as_recent_document_for_user(
            user=self._test_case_user
        )
        self._test_document.delete()

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_recently_accessed_document_list_view()
        self.assertNotContains(
            response=response, text=self._test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='File caching'), name='file_caching'
)

permission_cache_partition_purge = namespace.add_permission(
    label=_(message='Purge an object cache'),
    name='file_caching_cache_partition_purge'
)
permission_cache_purge = namespace.add_permission(
    label=_(message='Purge a file cache'), name='file_caching_cache_purge'
)
permission_cache_view = namespace.add_permission(
    label=_(message='View a file cache'), name='file_caching_cache_view'
)



import os
from pathlib import Path
import sys

from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.literals import COMMAND_NAME_SETTINGS_REVERT
from mayan.apps.smart_settings.utils import SettingNamespaceSingleton

from .literals import DEFAULT_SECRET_KEY, SECRET_KEY_FILENAME, SYSTEM_DIR

BASE_DIR = Path(__file__).resolve().parent.parent

setting_namespace = SettingNamespaceSingleton(
    global_symbol_table=globals()
)

if COMMAND_NAME_SETTINGS_REVERT in sys.argv:
    setting_namespace.update_globals(only_critical=True)
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': Path(MEDIA_ROOT, 'db.sqlite3')  # NOQA: F821
        }
    }
else:
    setting_namespace.update_globals()

try:
    SECRET_KEY = os.environ['MAYAN_SECRET_KEY']
except KeyError:
    path_secret_key = Path(
        MEDIA_ROOT, SYSTEM_DIR, SECRET_KEY_FILENAME  # NOQA: F821
    )
    try:
        with path_secret_key.open(mode='rb') as file_object:  # NOQA: F821
            SECRET_KEY = file_object.read().strip()
    except FileNotFoundError:
        SECRET_KEY = DEFAULT_SECRET_KEY

# Application definition

INSTALLED_APPS = (
    # Placed at the top so it can preload all events defined by apps.
    'mayan.apps.events.apps.EventsApp',
    # Placed at the top so it can override any template.
    'mayan.apps.appearance.apps.AppearanceApp',
    # Django
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.forms',
    'django.contrib.staticfiles',
    # 3rd party.
    'actstream',
    'corsheaders',
    'django_celery_beat',
    'formtools',
    'mathfilters',
    'mozilla_django_oidc',
    'mptt',
    'rest_framework',
    'rest_framework.authtoken',
    'solo',
    'stronghold',
    'widget_tweaks',
    # Base apps
    # Moved to the top to ensure Mayan app logging is initialized and
    # available as soon as possible.
    'mayan.apps.logging.apps.LoggingApp',
    # Task manager goes to the top to ensure all queues are created before any
    # other app tries to use them.
    'mayan.apps.task_manager.apps.TaskManagerApp',
    'mayan.apps.acls.apps.ACLsApp',
    # User management app must go before authentication to ensure the Group
    # and User models are properly setup using runtime methods.
    'mayan.apps.user_management.apps.UserManagementApp',
    'mayan.apps.authentication.apps.AuthenticationApp',
    'mayan.apps.authentication_oidc.apps.AuthenticationOIDCApp',
    'mayan.apps.authentication_otp.apps.AuthenticationOTPApp',
    'mayan.apps.autoadmin.apps.AutoAdminAppConfig',
    'mayan.apps.backends.apps.BackendsApp',
    'mayan.apps.common.apps.CommonApp',
    'mayan.apps.converter.apps.ConverterApp',
    'mayan.apps.credentials.apps.CredentialsApp',
    'mayan.apps.dashboards.apps.DashboardsApp',
    'mayan.apps.databases.apps.DatabasesApp',
    'mayan.apps.dependencies.apps.DependenciesApp',
    'mayan.apps.django_gpg.apps.DjangoGPGApp',
    'mayan.apps.dynamic_search.apps.DynamicSearchApp',
    'mayan.apps.file_caching.apps.FileCachingConfig',
    'mayan.apps.locales.apps.LocalesApp',
    'mayan.apps.lock_manager.apps.LockManagerApp',
    'mayan.apps.messaging.apps.MessagingApp',
    'mayan.apps.mime_types.apps.MIMETypesApp',
    'mayan.apps.navigation.apps.NavigationApp',
    'mayan.apps.organizations.apps.OrganizationsApp',
    'mayan.apps.permissions.apps.PermissionsApp',
    'mayan.apps.platform.apps.PlatformApp',
    'mayan.apps.quotas.apps.QuotasApp',
    'mayan.apps.rest_api.apps.RESTAPIApp',
    'mayan.apps.smart_settings.apps.SmartSettingsApp',
    'mayan.apps.storage.apps.StorageApp',
    'mayan.apps.templating.apps.TemplatingApp',
    'mayan.apps.testing.apps.TestingApp',
    'mayan.apps.views.apps.ViewsApp',
    # Obsolete apps. Need to remain to allow migrations to execute.
    'mayan.apps.announcements.apps.AnnouncementsApp',
    'mayan.apps.motd.apps.MOTDApp',
    # Document apps.
    'mayan.apps.cabinets.apps.CabinetsApp',
    'mayan.apps.checkouts.apps.CheckoutsApp',
    'mayan.apps.document_comments.apps.DocumentCommentsApp',
    'mayan.apps.document_downloads.apps.DocumentDownloadsApp',
    'mayan.apps.document_exports.apps.DocumentExportsApp',
    'mayan.apps.document_indexing.apps.DocumentIndexingApp',
    'mayan.apps.document_parsing.apps.DocumentParsingApp',
    'mayan.apps.document_signatures.apps.DocumentSignaturesApp',
    'mayan.apps.document_states.apps.DocumentStatesApp',
    'mayan.apps.documents.apps.DocumentsApp',
    'mayan.apps.duplicates.apps.DuplicatesApp',
    'mayan.apps.file_metadata.apps.FileMetadataApp',
    'mayan.apps.file_metadata_clamav.apps.FileMetadataClamAVApp',
    'mayan.apps.linking.apps.LinkingApp',
    'mayan.apps.mailer.apps.MailerApp',
    'mayan.apps.mayan_statistics.apps.StatisticsApp',
    'mayan.apps.metadata.apps.MetadataApp',
    'mayan.apps.mirroring.apps.MirroringApp',
    'mayan.apps.ocr.apps.OCRApp',
    'mayan.apps.redactions.apps.RedactionsApp',
    'mayan.apps.signature_captures.apps.SignatureCapturesApp',
    'mayan.apps.source_compressed.apps.SourceCompressedApp',
    'mayan.apps.source_interactive.apps.SourceInteractiveApp',
    'mayan.apps.source_periodic.apps.SourcePeriodicApp',
    'mayan.apps.source_emails.apps.SourceEmailsApp',
    'mayan.apps.source_sane_scanners.apps.SourceSaneScannersApp',
    'mayan.apps.source_staging_folders.apps.SourceStagingFoldersApp',
    'mayan.apps.source_staging_storages.apps.SourceStagingStorageApp',
    'mayan.apps.source_generated_files.apps.SourceGeneratedFileApp',
    'mayan.apps.source_stored_files.apps.SourceStoredFileApp',
    'mayan.apps.source_watch_folders.apps.SourceWatchFoldersApp',
    'mayan.apps.source_watch_storages.apps.SourceWatchStorageApp',
    'mayan.apps.source_web_forms.apps.SourceWebFormsApp',
    'mayan.apps.sources.apps.SourcesApp',
    'mayan.apps.tags.apps.TagsApp',
    'mayan.apps.web_links.apps.WebLinksApp',
    # Placed after rest_api to allow template overriding.
    'drf_yasg',
)

MIDDLEWARE = (
    'mayan.apps.logging.middleware.error_logging.ErrorLoggingMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'whitenoise.middleware.WhiteNoiseMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'mayan.apps.locales.middleware.locales.UserLocaleProfileMiddleware',
    'mayan.apps.authentication.middleware.impersonate.ImpersonateMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'stronghold.middleware.LoginRequiredMiddleware',
    'mayan.apps.views.middleware.ajax_redirect.AjaxRedirect'
)

ROOT_URLCONF = 'mayan.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages'
            ],
            'loaders': [
                'django.template.loaders.filesystem.Loader',
                'django.template.loaders.app_directories.Loader'
            ]
        }
    }
]

WSGI_APPLICATION = 'mayan.wsgi.application'

# Password validation

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator'
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator'
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator'
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator'
    }
]

# Internationalization

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)

STATIC_URL = '/static/'

# ------------ Custom settings section ----------

LANGUAGES = (
    ('ar-eg', _(message='Arabic (Egypt)')),
    ('ar', _(message='Arabic')),
    ('bg', _(message='Bulgarian')),
    ('bs', _(message='Bosnian')),
    ('ca', _(message='Catalan')),
    ('cs', _(message='Czech')),
    ('da', _(message='Danish')),
    ('de-at', _(message='German (Austria)')),
    ('de-de', _(message='German (Germany)')),
    ('el', _(message='Greek')),
    ('en', _(message='English')),
    ('es', _(message='Spanish')),
    ('es-mx', _(message='Spanish (Mexico)')),
    ('es-pr', _(message='Spanish (Puerto Rico)')),
    ('fa', _(message='Persian')),
    ('fr', _(message='French')),
    ('he-il', _(message='Hebrew (Israel)')),
    ('hu', _(message='Hungarian')),
    ('hr', _(message='Croatian')),
    ('id', _(message='Indonesian')),
    ('it', _(message='Italian')),
    ('lv', _(message='Latvian')),
    ('mn-mn', _(message='Mongolian (Mongolia)')),
    ('nl', _(message='Dutch')),
    ('pl', _(message='Polish')),
    ('pt', _(message='Portuguese')),
    ('pt-br', _(message='Portuguese (Brazil)')),
    ('ro-ro', _(message='Romanian (Romania)')),
    ('ru', _(message='Russian')),
    ('sl', _(message='Slovenian')),
    ('sq', _(message='Albanian')),
    ('th', _(message='Thai')),
    ('tr', _(message='Turkish')),
    ('tr-tr', _(message='Turkish (Turkey)')),
    ('uk', _(message='Ukrainian')),
    ('vi', _(message='Vietnamese')),
    ('zh-cn', _(message='Chinese (China)')),
    ('zh-hans', _(message='Chinese (Simplified)')),
    ('zh-tw', _(message='Chinese (Taiwan)'))
)

MEDIA_URL = 'media/'

SITE_ID = 1

STATIC_ROOT = os.environ.get(
    'MAYAN_STATIC_ROOT', Path(MEDIA_ROOT, 'static')  # NOQA: F821
)

MEDIA_URL = 'media/'

# Silence warning and keep default for the time being.
# https://docs.djangoproject.com/en/3.2/releases/3.2/#customizing-type-of-auto-created-primary-keys
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'mayan.apps.views.finders.MayanAppDirectoriesFinder'
)

STORAGES = {}
STORAGES['staticfiles'] = {
    'BACKEND': 'whitenoise.storage.CompressedManifestStaticFilesStorage'
}

TEST_RUNNER = 'mayan.apps.testing.runner.MayanTestRunner'

# ---------- Django REST framework -----------

REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework.authentication.TokenAuthentication',
        'rest_framework.authentication.BasicAuthentication'
    ),
    'DEFAULT_PAGINATION_CLASS': 'mayan.apps.rest_api.pagination.MayanPageNumberPagination',
    'EXCEPTION_HANDLER': 'mayan.apps.rest_api.exception_handlers.mayan_exception_handler'
}

# --------- Pagination --------

PAGINATION_SETTINGS = {
    'PAGE_RANGE_DISPLAYED': 5,
    'MARGIN_PAGES_DISPLAYED': 2
}

# ----------- Celery ----------

CELERY_ACCEPT_CONTENT = ('json',)
CELERY_BEAT_SCHEDULE = {}
CELERY_BEAT_SCHEDULER = 'django_celery_beat.schedulers.DatabaseScheduler'
CELERY_DISABLE_RATE_LIMITS = True
CELERY_ENABLE_UTC = True
CELERY_RESULT_SERIALIZER = 'json'
CELERY_TASK_ALWAYS_EAGER = False
CELERY_TASK_CREATE_MISSING_QUEUES = True
CELERY_TASK_DEFAULT_QUEUE = 'celery'
CELERY_TASK_EAGER_PROPAGATES = True
CELERY_TASK_QUEUES = []
CELERY_TASK_ROUTES = {}
CELERY_TASK_SERIALIZER = 'json'
CELERY_TIMEZONE = 'UTC'

# ------------ CORS ------------

CORS_ORIGIN_ALLOW_ALL = True

# ------ Timezone --------

TIMEZONE_COOKIE_NAME = 'django_timezone'
TIMEZONE_SESSION_KEY = 'django_timezone'

# ----- Stronghold -------

STRONGHOLD_PUBLIC_URLS = (r'^/favicon\.ico$',)

# ----- Swagger --------

SWAGGER_SETTINGS = {
    'DEFAULT_INFO': 'rest_api.schemas.openapi_info',
    'DEFAULT_MODEL_DEPTH': 1,
    'DOC_EXPANSION': 'None'
}

# ------ End -----

BASE_INSTALLED_APPS = INSTALLED_APPS

for app in INSTALLED_APPS:
    if 'mayan.apps.{}'.format(app) in BASE_INSTALLED_APPS:
        raise ImproperlyConfigured(
            'Update the app references in the file config.yml as detailed '
            'in https://docs.mayan-edms.com/releases/3.2.html#backward-incompatible-changes'
        )

repeated_apps = tuple(
    set(COMMON_EXTRA_APPS_PRE).intersection(  # NOQA: F821
        set(COMMON_EXTRA_APPS)  # NOQA: F821
    )
)
if repeated_apps:
    raise ImproperlyConfigured(
        'Apps "{}" cannot be specified in `COMMON_EXTRA_APPS_PRE` and '
        '`COMMON_EXTRA_APPS` at the same time.'.format(
            ', '.join(
                tuple(repeated_apps)
            )
        )
    )

INSTALLED_APPS = tuple(
    COMMON_EXTRA_APPS_PRE or ()  # NOQA: F821
) + INSTALLED_APPS

INSTALLED_APPS = INSTALLED_APPS + tuple(
    COMMON_EXTRA_APPS or ()  # NOQA: F821
)

INSTALLED_APPS = [
    APP for APP in INSTALLED_APPS if APP not in (
        COMMON_DISABLED_APPS or ()  # NOQA: F821
    )
]

if not DATABASES:
    if DATABASE_ENGINE:  # NOQA: F821
        DATABASES = {
            'default': {
                'ENGINE': DATABASE_ENGINE,  # NOQA: F821
                'NAME': DATABASE_NAME,  # NOQA: F821
                'USER': DATABASE_USER,  # NOQA: F821
                'PASSWORD': DATABASE_PASSWORD,  # NOQA: F821
                'HOST': DATABASE_HOST,  # NOQA: F821
                'PORT': DATABASE_PORT,  # NOQA: F821
                'CONN_MAX_AGE': DATABASE_CONN_MAX_AGE  # NOQA: F821
            }
        }
    else:
        DATABASES = {
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': str(
                    Path(MEDIA_ROOT, 'db.sqlite3')  # NOQA: F821
                )
            }
        }



from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_web_link_created, event_web_link_edited
from ..links import link_web_link_instance_view
from ..models import ResolvedWebLink, WebLink
from ..permissions import (
    permission_web_link_create, permission_web_link_delete,
    permission_web_link_edit, permission_web_link_instance_view,
    permission_web_link_view
)

from .literals import TEST_WEB_LINK_TEMPLATE
from .mixins import (
    DocumentTypeAddRemoveWebLinkViewTestMixin,
    WebLinkDocumentTypeViewTestMixin, WebLinkViewTestMixin
)


class DocumentTypeAddRemoveWebLinkViewTestCase(
    DocumentTypeAddRemoveWebLinkViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_web_link()

    def test_document_type_web_link_add_remove_get_view_no_permission(self):
        self._test_document_type.web_links.add(self._test_web_link)

        self._clear_events()

        response = self._request_test_document_type_web_link_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_web_link),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_add_remove_get_view_with_document_type_access(self):
        self._test_document_type.web_links.add(self._test_web_link)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_web_link_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self._test_document_type),
            status_code=200
        )
        self.assertNotContains(
            response=response, text=str(self._test_web_link),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_add_remove_get_view_with_web_link_access(self):
        self._test_document_type.web_links.add(self._test_web_link)

        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_web_link_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_web_link),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_add_remove_get_view_with_full_access(self):
        self._test_document_type.web_links.add(self._test_web_link)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_web_link_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self._test_document_type),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self._test_web_link),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_add_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_type_web_link_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_web_link not in self._test_document_type.web_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_add_view_with_document_type_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_web_link_add_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_web_link not in self._test_document_type.web_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_add_view_with_web_link_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_web_link_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_web_link not in self._test_document_type.web_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_add_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_web_link_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_web_link in self._test_document_type.web_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_web_link)
        self.assertEqual(events[0].verb, event_web_link_edited.id)

    def test_document_type_web_link_remove_view_no_permission(self):
        self._test_document_type.web_links.add(self._test_web_link)

        self._clear_events()

        response = self._request_test_document_type_web_link_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_web_link in self._test_document_type.web_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_remove_view_with_document_type_access(self):
        self._test_document_type.web_links.add(self._test_web_link)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_web_link_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_web_link in self._test_document_type.web_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_remove_view_with_web_link_access(self):
        self._test_document_type.web_links.add(self._test_web_link)

        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_web_link_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_web_link in self._test_document_type.web_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_web_link_remove_view_with_full_access(self):
        self._test_document_type.web_links.add(self._test_web_link)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_web_link_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_web_link not in self._test_document_type.web_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_web_link)
        self.assertEqual(events[0].verb, event_web_link_edited.id)


class WebLinkViewTestCase(WebLinkViewTestMixin, GenericViewTestCase):
    def test_web_link_create_view_no_permission(self):
        web_link_count = WebLink.objects.count()

        self._clear_events()

        response = self._request_test_web_link_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(WebLink.objects.count(), web_link_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_create_view_with_permissions(self):
        self.grant_permission(permission=permission_web_link_create)

        web_link_count = WebLink.objects.count()

        self._clear_events()

        response = self._request_test_web_link_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(WebLink.objects.count(), web_link_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_web_link)
        self.assertEqual(events[0].verb, event_web_link_created.id)

    def test_web_link_delete_view_no_permission(self):
        self._create_test_web_link()

        web_link_count = WebLink.objects.count()

        self._clear_events()

        response = self._request_test_web_link_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(WebLink.objects.count(), web_link_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_delete_view_with_access(self):
        self._create_test_web_link()

        self.grant_access(
            obj=self._test_web_link, permission=permission_web_link_delete
        )

        web_link_count = WebLink.objects.count()

        self._clear_events()

        response = self._request_test_web_link_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(WebLink.objects.count(), web_link_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_edit_view_no_permission(self):
        self._create_test_web_link()

        web_link_label = self._test_web_link.label

        self._clear_events()

        response = self._request_test_web_link_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_web_link.refresh_from_db()
        self.assertEqual(self._test_web_link.label, web_link_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_edit_view_with_access(self):
        self._create_test_web_link()

        self.grant_access(
            obj=self._test_web_link, permission=permission_web_link_edit
        )

        web_link_label = self._test_web_link.label

        self._clear_events()

        response = self._request_test_web_link_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_web_link.refresh_from_db()
        self.assertNotEqual(self._test_web_link.label, web_link_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_web_link)
        self.assertEqual(events[0].verb, event_web_link_edited.id)

    def test_web_link_list_view_with_no_permission(self):
        self._create_test_web_link()

        self._clear_events()

        response = self._request_test_web_link_list_view()
        self.assertNotContains(
            response=response, text=self._test_web_link.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_list_view_with_access(self):
        self._create_test_web_link()

        self.grant_access(obj=self._test_web_link, permission=permission_web_link_view)

        self._clear_events()

        response = self._request_test_web_link_list_view()
        self.assertContains(
            response=response, text=self._test_web_link.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class WebLinkDocumentTypeViewTestCase(
    WebLinkDocumentTypeViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_web_link()

    def test_web_link_document_type_add_remove_get_view_no_permission(self):
        self._test_web_link.document_types.add(self._test_document_type)

        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self._clear_events()

        response = self._request_test_web_link_document_type_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_web_link),
            status_code=404
        )

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_remove_get_view_with_document_type_access(self):
        self._test_web_link.document_types.add(self._test_document_type)

        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_web_link),
            status_code=404
        )

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_remove_get_view_with_web_link_access(self):
        self._test_web_link.document_types.add(self._test_document_type)

        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self._test_web_link),
            status_code=200
        )

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_remove_get_view_with_full_access(self):
        self._test_web_link.document_types.add(self._test_document_type)

        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self._test_document_type),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self._test_web_link),
            status_code=200
        )

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_view_no_permission(self):
        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_document_type, permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_view_with_document_type_access(self):
        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_document_type, permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_view_with_web_link_access(self):
        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_web_link, permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_add_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_view_with_full_access(self):
        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_web_link, permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_web_link)
        self.assertEqual(events[0].verb, event_web_link_edited.id)

    def test_web_link_document_type_remove_view_no_permission(self):
        self._test_web_link.document_types.add(self._test_document_type)

        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_document_type, permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_remove_view_with_document_type_access(self):
        self._test_web_link.document_types.add(self._test_document_type)

        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_document_type, permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_remove_view_with_web_link_access(self):
        self._test_web_link.document_types.add(self._test_document_type)

        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_web_link, permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_remove_view_with_full_access(self):
        self._test_web_link.document_types.add(self._test_document_type)

        test_web_link_document_type_count = self._test_web_link.document_types.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_web_link, permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_web_link.document_types.count(),
            test_web_link_document_type_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_web_link)
        self.assertEqual(events[0].verb, event_web_link_edited.id)


class DocumentWebLinkViewTestCase(
    WebLinkViewTestMixin, GenericDocumentViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_web_link(add_test_document_type=True)

    def test_document_web_links_list_view_no_permission(self):
        response = self._request_test_document_web_link_list_view()
        self.assertNotContains(
            response=response, text=str(self._test_document),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_web_link),
            status_code=404
        )

    def test_document_web_links_list_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )

        response = self._request_test_document_web_link_list_view()
        self.assertContains(
            response=response, text=str(self._test_document),
            status_code=200
        )
        self.assertNotContains(
            response=response, text=str(self._test_web_link),
            status_code=200
        )

    def test_document_web_links_list_view_with_web_link_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        response = self._request_test_document_web_link_list_view()
        self.assertNotContains(
            response=response, text=str(self._test_document),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_web_link),
            status_code=404
        )

    def test_document_web_links_list_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        response = self._request_test_document_web_link_list_view()
        self.assertContains(
            response=response, text=str(self._test_document),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self._test_web_link),
            status_code=200
        )
        # Test if a valid resolved link navigate link is present.
        # Ensures the view is returning ResolvedWebLink proxies and not the
        # base model WebLink.
        context = self._get_context_from_test_response(response=response)
        context['object'] = ResolvedWebLink.objects.get_for(
            document=self._test_document, user=self._test_case_user
        ).first()
        resolved_web_link_link = link_web_link_instance_view.resolve(context=context)
        self.assertContains(
            response=response, text=resolved_web_link_link.url,
            status_code=200
        )

    def test_trashed_document_web_links_list_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        self._test_document.delete()

        response = self._request_test_document_web_link_list_view()
        self.assertEqual(response.status_code, 404)

    def test_document_resolved_web_link_view_no_permission(self):
        response = self._request_test_document_web_link_instance_view()
        self.assertEqual(response.status_code, 404)

    def test_document_resolved_web_link_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )

        response = self._request_test_document_web_link_instance_view()
        self.assertEqual(response.status_code, 404)

    def test_document_resolved_web_link_view_with_web_link_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        response = self._request_test_document_web_link_instance_view()
        self.assertEqual(response.status_code, 404)

    def test_document_resolved_web_link_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        response = self._request_test_document_web_link_instance_view()
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, TEST_WEB_LINK_TEMPLATE.replace(
                '{{ document.uuid }}', str(self._test_document.uuid)
            )
        )

    def test_trashed_document_resolved_web_link_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        self._test_document.delete()

        response = self._request_test_document_web_link_instance_view()
        self.assertEqual(response.status_code, 404)
